"""
This module contains utilities to push a model to the hub and pull from the
hub.
"""
from __future__ import annotations

import collections
import itertools
import json
import os
import shutil
from pathlib import Path
from typing import Any, List, Literal, MutableMapping, Optional, Sequence, Union

import numpy as np
from huggingface_hub import HfApi, InferenceApi, snapshot_download
from sklearn.utils import check_array

SUPPORTED_TASKS = [
    "tabular-classification",
    "tabular-regression",
    "text-classification",
    "text-regression",
]


def _validate_folder(path: Union[str, Path]) -> None:
    """Validate the contents of a folder.

    This function checks if the contents of a folder make a valid repo for a
    scikit-learn based repo on the Hugging Face Hub.

    A valid repository is one which is understood by the Hub as well as this
    library to run and use the model. Otherwise anything can be put as a model
    repository on the Hub and use it as a `git` and `git lfs` server.

    Raises a ``TypeError`` if invalid.

    Parameters
    ----------
    path: str or Path
        The location of the repo.

    Raises
    ------
    TypeError
        Raised when the passed path is invalid.

    Returns
    -------
    None
    """
    path = Path(path)
    if not path.is_dir():
        raise TypeError("The given path is not a directory.")

    config_path = path / "config.json"
    if not config_path.exists():
        raise TypeError("Configuration file `config.json` missing.")

    with open(config_path, "r") as f:
        config = json.load(f)

    model_path = config.get("sklearn", {}).get("model", {}).get("file", None)
    if not model_path:
        raise TypeError(
            "Model file not configured in the configuration file. It should be stored"
            " in the hf_hub.sklearn.model key."
        )

    if not (path / model_path).exists():
        raise TypeError(f"Model file {model_path} does not exist.")


def _get_example_input_from_tabular_data(data):
    """Returns the example input of a model for a tabular task.

    The input is converted into a dictionary which is then stored in the config
    file.

    Parameters
    ----------
    data: array-like
        The input needs to be either a ``pandas.DataFrame``, a 2D
        ``numpy.ndarray`` or a list/tuple that can be converted to a 2D
        ``numpy.ndarray``. The first 3 rows are used as example input.

    Returns
    -------
    example_input: dict of lists
        The example input of the model as accepted by Hugging Face's backend.
    """
    try:
        import pandas as pd

        if isinstance(data, pd.DataFrame):
            return {x: data[x][:3].to_list() for x in data.columns}
    except ImportError:
        # pandas is not installed, the data cannot be a pandas DataFrame
        pass

    # here we convert the first three rows of `data` to a dict of lists
    # to be stored in the config file
    if isinstance(data, (np.ndarray, list, tuple)):
        data_slice = data[:3]
        # This will raise a ValueError if the array is not 2D
        data_slice_array = check_array(data_slice, ensure_2d=True)
        return {
            f"x{x}": data_slice_array[:, x].tolist()
            for x in range(data_slice_array.shape[1])
        }

    raise ValueError(
        "The data is not a pandas.DataFrame, a 2D numpy.ndarray or a "
        "list/tuple that can be converted to a 2D numpy.ndarray."
    )


def _get_example_input_from_text_data(data: Sequence[str]):
    """Returns the example input of a model for a text task.

    The input is converted into a dictionary which is then stored in the config
    file.

    Parameters
    ----------
    data: Sequence[str]
        A sequence of strings. The first 3 elements are used as example input.

    Returns
    -------
    example_input: dict of lists
        The example input of the model as accepted by Hugging Face's backend.
    """

    def _head(data, n):
        is_data_subscriptable = hasattr(data, "__getitem__")
        if is_data_subscriptable:
            return data[:n]

        return list(itertools.islice(data, n))

    def _is_sequence_of_strings(data):
        return not isinstance(data, str) and all(isinstance(x, str) for x in data)

    error_message = "The data needs to be a sequence of strings."
    try:
        data_head = _head(data, n=3)
        if _is_sequence_of_strings(data_head):
            return {"data": data_head}
        else:
            raise ValueError(error_message)
    except TypeError as e:
        raise ValueError(error_message) from e


def _get_column_names(data):
    """Returns the column names of the input.

    If data is not a ``pandas.DataFrame``, column names are assumed to be
    ``x0`` to ``xn-1``, where ``n`` is the number of columns.

    Parameters
    ----------
    data: array-like
        The data whose columns names are to be returned. Must be a
        ``pandas.DataFrame``, a 2D ``numpy.ndarray`` or a list/tuple that can
        be converted to a 2D ``numpy.ndarray``.

    Returns
    -------
    columns: list of strings
        A list of strings. Each string is a column name.
    """
    try:
        import pandas as pd

        if isinstance(data, pd.DataFrame):
            return list(data.columns)
    except ImportError:
        # pandas is not installed, the data cannot be a pandas DataFrame
        pass

    # TODO: this is going to fail for Structured Arrays. We can add support for
    # them later if we see need for it.
    if isinstance(data, (np.ndarray, list, tuple)):
        # This will raise a ValueError if the array is not 2D
        data_array = check_array(data, ensure_2d=True)
        return [f"x{x}" for x in range(data_array.shape[1])]

    raise ValueError(
        "The data is not a pandas.DataFrame, a 2D numpy.ndarray or a "
        "list/tuple that can be converted to a 2D numpy.ndarray."
    )


def _create_config(
    *,
    model_path: Union[str, Path],
    requirements: List[str],
    dst: Union[str, Path],
    task: Literal[
        "tabular-classification",
        "tabular-regression",
        "text-classification",
        "text-regression",
    ],
    data,
    model_format: Literal[
        "skops",
        "pickle",
        "auto",
    ] = "auto",
    use_intelex: bool = False,
) -> None:
    """Write the configuration into a ``config.json`` file.

    Parameters
    ----------
    model_path : str, or Path
        The relative path (from the repo root) to the model file.

    requirements : list of str
        A list of required packages. The versions are then extracted from the
        current environment.

    dst : str, or Path
        The path to an existing folder where the config file should be created.

    task: "tabular-classification", "tabular-regression",
    "text-classification", /
            or "text-regression"
        The task of the model, which determines the input and output type of
        the model. It can be one of: ``tabular-classification``,
        ``tabular-regression``, ``text-classification``, ``text-regression``.

    data: array-like or sequence
        The input to the model. This is used for two purposes:

            1. Save an example input to the model, which is used by
               HuggingFace's backend and shown in the widget of the model's
               page.
            2. Store the columns and their order of the input, which is used by
               HuggingFace's backend to pass the data in the right form to the
               model.

        The first 3 input values are used as example inputs. If the task is
        ``tabular-classification`` or ``tabular-regression``, then data is
        expected to be an array-like. Otherwise, it is expected to be an
        sequence of strings.

    model_format: str (default="auto")
        The format used to persist the model. Can be ``"auto"``, ``"skops"``
        or ``"pickle"``. Defaults to ``"auto"``, which would mean:

        - ``"pickle"`` if the extension is one of ``{".pickle", ".pkl", ".joblib"}``
        - ``"skops"`` if the extension is ``".skops"``

    use_intelex: bool (default=False)
        Whether to enable ``scikit-learn-intelex``. This can accelerate some
        sklearn models by a large factor with the right hardware. In most cases,
        enabling this option should not break any code, even if the model was
        not initially trained with scikit-learn intelex and even if the hardware
        does not support it. For more info, see
        https://intel.github.io/scikit-learn-intelex/.
    """

    # so that we don't have to explicitly add keys and they're added as a
    # dictionary if they are not found
    # see: https://stackoverflow.com/a/13151294/2536294
    def recursively_default_dict() -> MutableMapping:
        return collections.defaultdict(recursively_default_dict)

    if model_format == "auto":
        extension = Path(model_path).suffix
        if extension in [".pkl", ".pickle", ".joblib"]:
            model_format = "pickle"
        elif extension == ".skops":
            model_format = "skops"
    if model_format not in ["skops", "pickle"]:
        raise ValueError(
            "Cannot determine the input file format. Please indicate the format using"
            " the `model_format` argument."
        )

    config = recursively_default_dict()
    config["sklearn"]["model"]["file"] = str(model_path)
    config["sklearn"]["environment"] = requirements
    config["sklearn"]["task"] = task
    config["sklearn"]["model_format"] = model_format
    config["sklearn"]["use_intelex"] = use_intelex

    if "tabular" in task:
        config["sklearn"]["example_input"] = _get_example_input_from_tabular_data(data)
        config["sklearn"]["columns"] = _get_column_names(data)
    elif "text" in task:
        config["sklearn"]["example_input"] = _get_example_input_from_text_data(data)

    dump_json(Path(dst) / "config.json", config)


def _check_model_file(path: str | Path) -> Path:
    """Perform sanity checks on the model file

    Parameters
    ----------
    path : str or Path
      The model path

    Returns
    -------
    path : Path
      The model path as a  ``pathlib.Path``.

    Raises
    ------
    OSError
      If the model file does not exist.
    """
    if not os.path.exists(path):
        raise OSError(f"Model file '{path}' does not exist.")

    if os.path.getsize(path) == 0:
        raise RuntimeError(f"Model file '{path}' is empty.")

    return Path(path)


def init(
    *,
    model: Union[str, Path],
    requirements: List[str],
    dst: Union[str, Path],
    task: Literal[
        "tabular-classification",
        "tabular-regression",
        "text-classification",
        "text-regression",
    ],
    data,
    model_format: Literal[
        "skops",
        "pickle",
        "auto",
    ] = "auto",
    use_intelex: bool = False,
) -> None:
    """Initialize a scikit-learn based Hugging Face repo.

    Given a pickled model and a set of required packages, this function
    initializes a folder to be a valid Hugging Face scikit-learn based repo.

    Parameters
    ----------
    model: str, or Path
        The path to a model pickle file.

    requirements: list of str
        A list of required packages. The versions are then extracted from the
        current environment.

    dst: str, or Path
        The path to a non-existing or empty folder which is to be initialized.

    task: str
        The task of the model, which determines the input and output type of
        the model. It can be one of: ``tabular-classification``,
        ``tabular-regression``, ``text-classification``, ``text-regression``.

    data: array-like
        The input to the model. This is used for two purposes:

            1. Save an example input to the model, which is used by
               HuggingFace's backend and shown in the widget of the model's
               page.
            2. Store the columns and their order of the input, which is used by
               HuggingFace's backend to pass the data in the right form to the
               model.

        The first 3 input values are used as example inputs.

        If ``task`` is ``"tabular-classification"`` or ``"tabular-regression"``,
        the data needs to be a :class:`pandas.DataFrame` or a
        :class:`numpy.ndarray`. If ``task`` is ``"text-classification"`` or
        ``"text-regression"``, the data needs to be a ``list`` of strings.

    model_format: str (default="auto")
        The format the model was persisted in. Can be ``"auto"``, ``"skops"``
        or ``"pickle"``. Defaults to ``"auto"`` that relies on file extension.

    use_intelex: bool (default=False)
        Whether to enable ``scikit-learn-intelex``. This can accelerate some
        sklearn models by a large factor with the right hardware. In most cases,
        enabling this option should not break any code, even if the model was
        not initially trained with scikit-learn intelex and even if the hardware
        does not support it. For more info, see
        https://intel.github.io/scikit-learn-intelex/.
    """
    dst = Path(dst)
    if dst.exists() and bool(next(dst.iterdir(), None)):
        raise OSError("None-empty dst path already exists!")

    if task not in SUPPORTED_TASKS:
        raise ValueError(
            f"Task {task} not supported. Supported tasks are: {SUPPORTED_TASKS}"
        )

    model = _check_model_file(model)

    dst.mkdir(parents=True, exist_ok=True)

    # add intelex requirement, if it's used and not already in requirements
    if use_intelex and not any(
        r.startswith("scikit-learn-intelex") for r in requirements
    ):
        requirements.append("scikit-learn-intelex")

    try:
        shutil.copy2(src=model, dst=dst)

        model_name = model.name
        _create_config(
            model_path=model_name,
            requirements=requirements,
            dst=dst,
            task=task,
            data=data,
            model_format=model_format,
            use_intelex=use_intelex,
        )
    except Exception:
        shutil.rmtree(dst)
        raise


def add_files(*files: str | Path, dst: str | Path, exist_ok: bool = False) -> None:
    """Add files to initialized repo.

    After having called :func:`.hub_utils.init`, use this function to add
    arbitrary files to be uploaded in addition to the model and model card.

    In particular, it can be useful to upload the script itself that produces
    those artifacts by calling ``hub_utils.add_files([__file__], dst=...)``.

    Parameters
    ----------
    *files : str or Path
        The files to be added.

    dst : str or Path
        Path to the initialized repo, same as used during
        :func:`.hub_utils.init`.

    exist_ok : bool (default=False)
        Whether it's okay or not to add a file that already exists. If
        ``True``, override the files, otherwise raise a ``FileExistsError``.

    Raises
    ------
    FileNotFoundError
        When the target folder or the files to be added are not found.

    FileExistsError
        When a file is added that already exists at the target location and
        ``exist_ok=False``.

    """
    dst = Path(dst)
    # check dst exists
    if not dst.exists():
        msg = f"Could not find '{dst}', did you run 'skops.hub_utils.init' first?"
        raise FileNotFoundError(msg)

    src_files = [Path(file) for file in files]
    # check that source files exist
    for file in src_files:
        if not file.exists():
            msg = f"File '{file}' could not be found."
            raise FileNotFoundError(msg)

    dst_files = [dst / Path(file).name for file in files]
    for src_file, dst_file in zip(src_files, dst_files):
        # check if target file already exists
        if dst_file.exists() and not exist_ok:
            msg = f"File '{src_file.name}' already found at '{dst}'."
            raise FileExistsError(msg)

        shutil.copy2(src_file, dst_file)


def dump_json(path, content):
    with open(Path(path), mode="w") as f:
        json.dump(content, f, sort_keys=True, indent=4)


def update_env(
    *, path: Union[str, Path], requirements: Union[List[str], None] = None
) -> None:
    """Update the environment requirements of a repo.

    This function takes the path to the repo, and updates the requirements of
    running the scikit-learn based model in the repo.

    Parameters
    ----------
    path: str, or Path
        The path to an existing local repo.

    requirements: list of str, optional
        The list of required packages for the model. If none is passed, the
        list of existing requirements is used and their versions are updated.

    """

    with open(Path(path) / "config.json") as f:
        config = json.load(f)

    config["sklearn"]["environment"] = requirements

    dump_json(Path(path) / "config.json", config)


def push(
    *,
    repo_id: str,
    source: Union[str, Path],
    token: str | None = None,
    commit_message: str | None = None,
    create_remote: bool = False,
    private: bool | None = None,
) -> None:
    """Pushes the contents of a model repo to Hugging Face Hub.

    This function validates the contents of the folder before pushing it to the
    Hub.

    Parameters
    ----------
    repo_id: str
        The ID of the destination repository in the form of ``OWNER/REPO_NAME``.

    source: str or Path
        A folder where the contents of the model repo are located.

    token: str, optional
        A token to push to the Hub. If not provided, the user should be already
        logged in using ``huggingface-cli login``.

    commit_message: str, optional
        The commit message to be used when pushing to the repo.

    create_remote: bool, default=False
        Whether to create the remote repository if it doesn't exist. If the
        remote repository doesn't exist and this parameter is ``False``, it
        raises an error. Otherwise it checks if the remote repository exists,
        and would create it if it doesn't.

    private: bool, default=None
        Whether the remote repository should be public or private. If ``True``
        or ``False`` is passed, this method will set the private/public status
        of the remote repository, regardless of it already existing or not. If
        ``None``, no change is applied.

        .. versionadded:: 0.3

    Returns
    -------
    None

    Raises
    ------
    TypeError
        This function raises a ``TypeError`` if the contents of the source
        folder do not make a valid Hugging Face Hub scikit-learn based repo.
    """
    _validate_folder(path=source)
    client = HfApi()

    if create_remote:
        client.create_repo(
            repo_id=repo_id, token=token, repo_type="model", exist_ok=True
        )

    if private is not None:
        client.update_repo_visibility(repo_id=repo_id, private=private, token=token)

    client.upload_folder(
        repo_id=repo_id,
        path_in_repo=".",
        folder_path=source,
        commit_message=commit_message,
        commit_description=None,
        token=token,
        repo_type=None,
        revision=None,
        create_pr=False,
    )


def get_config(path: Union[str, Path]) -> dict[str, Any]:
    """Returns the configuration of a project.

    Parameters
    ----------
    path: str
        The path to the directory holding the project and its ``config.json``
        configuration file.

    Returns
    -------
    config: dict
        A dictionary which holds the configs of the project.
    """
    with open(Path(path) / "config.json", "r") as f:
        config = json.load(f)
    return config


def get_requirements(path: Union[str, Path]) -> List[str]:
    """Returns the requirements of a project.

    Parameters
    ----------
    path: str
        The path to the director holding the project and its ``config.json``
        configuration file.

    Returns
    -------
    requirements: list of str
        The list of requirements which can be passed to the package manager to
        be installed.
    """
    config = get_config(path)
    return config.get("sklearn", dict()).get("environment", list())


def download(
    *,
    repo_id: str,
    dst: Union[str, Path],
    revision: str | None = None,
    token: str | None = None,
    keep_cache: bool = True,
    **kwargs: Any,
) -> None:
    """Download a repository into a directory.

    The directory needs to be an empty or a non-existing one.

    Parameters
    ----------
    repo_id: str
        The ID of the Hugging Face Hub repository in the form of
        ``OWNER/REPO_NAME``.

    dst: str, or Path
        The directory to which the files are downloaded.

    revision: str, optional
        The revision of the project to download. This can be a git tag, branch,
        or a git commit hash. By default the latest revision of the default
        branch is downloaded.

    token: str, optional
        The token to be used to download the files. Only required if the
        repository is private.

    keep_cache: bool, default=True
        Whether the cached data should be kept or removed after download. By
        default a copy of the cached files will be created in the ``dst``
        folder. If ``False``, the cache will be removed after the contents are
        copied. Note that the cache is git based and by default new files are
        only downloaded if there is a new revision of them on the hub. If you
        keep the cache, the old files are not removed after downloading the
        newer versions of them.

    kwargs: dict
        Other parameters to be passed to
        :func:`huggingface_hub.snapshot_download`.

    Returns
    -------
    None
    """
    dst = Path(dst)
    if dst.exists() and bool(next(dst.iterdir(), None)):
        raise OSError("None-empty dst path already exists!")

    # remove the folder only if it's empty and it exists
    if dst.exists():
        dst.rmdir()

    # TODO: Switch from use_auth_token to token once huggingface_hub<0.11 is
    # dropped. Until then, we ignore the mypy type check, because mypy doesn't
    # see that use_auth_token is handled by the decorator of snapshot_download.
    cached_folder = snapshot_download(
        repo_id=repo_id, revision=revision, use_auth_token=token, **kwargs  # type: ignore
    )
    shutil.copytree(cached_folder, dst)
    if not keep_cache:
        shutil.rmtree(path=cached_folder)


def get_model_output(repo_id: str, data: Any, token: Optional[str] = None) -> Any:
    """Returns the output of the model using Hugging Face Hub's inference API.

    See the :ref:`User Guide <hf_hub_inference>` for more details.

    Parameters
    ----------
    repo_id: str
        The ID of the Hugging Face Hub repository in the form of
        ``OWNER/REPO_NAME``.

    data: Any
        The input to be given to the model. This can be a
        :class:`pandas.DataFrame` or a :class:`numpy.ndarray`. If possible, you
        should always pass a :class:`pandas.DataFrame` with correct column
        names.

    token: str, optional
        The token to be used to call the inference API. Only required if the
        repository is private.

    Returns
    -------
    output: numpy.ndarray
        The output of the model.

    Notes
    -----
    If there are warnings or exceptions during inference, this function raises
    a :class:`RuntimeError` including the original errors and warnings
    returned from the server.

    Also note that if the model repo is private, the inference API would not be
    available.
    """
    # TODO: the "type: ignore" should eventually become unncessary when hf_hub
    # is updated
    model_info = HfApi().model_info(repo_id=repo_id, use_auth_token=token)  # type: ignore
    if not model_info.pipeline_tag:
        raise ValueError(
            f"Repo {repo_id} has no pipeline tag. You should set a valid 'task' in"
            " config.json and README.md files. This is automatically done for you if"
            " you pass a valid task to the skops.hub_utils.init() and"
            " skops.card.metadata_from_util() functions to generate those files."
        )

    try:
        inputs = {"data": data.to_dict(orient="list")}
    except AttributeError:
        # the input is not a pandas DataFrame
        inputs = {f"x{i}": data[:, i] for i in range(data.shape[1])}
        inputs = {"data": inputs}

    res = InferenceApi(repo_id=repo_id, task=model_info.pipeline_tag, token=token)(
        inputs=inputs
    )

    if isinstance(res, list):
        return np.array(res)
    else:
        raise RuntimeError(f"There were errors or warnings during inference: {res}")



import os
import pickle
import re
import tempfile
import textwrap
from pathlib import Path

import numpy as np
import pytest
import sklearn
from huggingface_hub import ModelCardData, metadata_load
from sklearn.datasets import load_iris
from sklearn.inspection import permutation_importance
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.metrics import f1_score, make_scorer
from sklearn.neighbors import KNeighborsClassifier

from skops import hub_utils
from skops.card import Card, metadata_from_config
from skops.card._model_card import (
    CONTENT_PLACEHOLDER,
    SKOPS_TEMPLATE,
    PlotSection,
    Section,
    TableSection,
    _load_model,
)
from skops.io import dump
from skops.utils.importutils import import_or_raise


def fit_model():
    X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
    y = np.dot(X, np.array([1, 2])) + 3
    reg = LinearRegression().fit(X, y)
    return reg


def save_model_to_file(model_instance, suffix):
    save_file_handle, save_file = tempfile.mkstemp(suffix=suffix, prefix="skops-test")
    if suffix in (".pkl", ".pickle"):
        with open(save_file, "wb") as f:
            pickle.dump(model_instance, f)
    elif suffix == ".skops":
        dump(model_instance, save_file)
    return save_file_handle, save_file


@pytest.mark.parametrize("suffix", [".pkl", ".pickle", ".skops"])
def test_load_model(suffix):
    model0 = LinearRegression(n_jobs=123)
    _, save_file = save_model_to_file(model0, suffix)
    loaded_model_str = _load_model(save_file, trusted=True)
    save_file_path = Path(save_file)
    loaded_model_path = _load_model(save_file_path, trusted=True)
    loaded_model_instance = _load_model(model0, trusted=True)

    assert loaded_model_str.n_jobs == 123
    assert loaded_model_path.n_jobs == 123
    assert loaded_model_instance.n_jobs == 123


@pytest.fixture
def model_card(model_diagram=True):
    model = fit_model()
    card = Card(model, model_diagram)
    yield card


@pytest.fixture
def iris_data():
    X, y = load_iris(return_X_y=True, as_frame=True)
    yield X, y


@pytest.fixture
def iris_estimator(iris_data):
    X, y = iris_data
    est = LogisticRegression(solver="liblinear").fit(X, y)
    yield est


@pytest.fixture
def iris_pkl_file(iris_estimator):
    pkl_file = tempfile.mkstemp(suffix=".pkl", prefix="skops-test")[1]
    with open(pkl_file, "wb") as f:
        pickle.dump(iris_estimator, f)
    yield pkl_file


@pytest.fixture
def iris_skops_file(iris_estimator):
    skops_folder = tempfile.mkdtemp()
    model_name = "model.skops"
    skops_path = Path(skops_folder) / model_name
    dump(iris_estimator, skops_path)
    yield skops_path


def _create_model_card_from_saved_model(
    destination_path,
    iris_estimator,
    iris_data,
    save_file,
):
    X, y = iris_data
    hub_utils.init(
        model=save_file,
        requirements=[f"scikit-learn=={sklearn.__version__}"],
        dst=destination_path,
        task="tabular-classification",
        data=X,
    )
    card = Card(iris_estimator, metadata=metadata_from_config(destination_path))
    card.save(Path(destination_path) / "README.md")
    return card


@pytest.fixture
def skops_model_card_metadata_from_config(
    destination_path, iris_estimator, iris_skops_file, iris_data
):
    yield _create_model_card_from_saved_model(
        destination_path, iris_estimator, iris_data, iris_skops_file
    )


@pytest.fixture
def pkl_model_card_metadata_from_config(
    destination_path, iris_estimator, iris_pkl_file, iris_data
):
    yield _create_model_card_from_saved_model(
        destination_path, iris_estimator, iris_data, iris_pkl_file
    )


@pytest.fixture
def destination_path():
    with tempfile.TemporaryDirectory(prefix="skops-test") as dir_path:
        yield Path(dir_path)


def test_save_model_card(destination_path, model_card):
    model_card.save(Path(destination_path) / "README.md")
    assert (Path(destination_path) / "README.md").exists()


CUSTOM_TEMPLATES = [None, {}, {"A Title", "Another Title", "A Title/A Section"}]  # type: ignore


class TestAddModelPlot:
    """Tests for the sklearn model repr"""

    def test_default(self, model_card):
        result = model_card.select(
            "Model description/Training Procedure/Model Plot"
        ).format()
        # don't compare whole text, as it's quite long and non-deterministic
        assert result.startswith("<style>#sk-")
        assert "<style>" in result
        assert result.endswith(
            "<pre>LinearRegression()</pre></div></div></div></div></div>"
        )

    def test_no_overflow(self, model_card):
        result = model_card.select(
            "Model description/Training Procedure/Model Plot"
        ).format()
        # test if the model doesn't overflow the huggingface models page
        assert result.count("sk-top-container") == 1
        assert 'style="overflow: auto;' in result

    def test_model_diagram_false(self):
        model = fit_model()
        model_card = Card(model, model_diagram=False)
        result = model_card.select(
            "Model description/Training Procedure/Model Plot"
        ).format()
        assert result == CONTENT_PLACEHOLDER

    def test_model_diagram_str(self):
        # if passing a str, use that as the section name
        model = fit_model()
        other_section_name = "Here is the model diagram"
        model_card = Card(model, model_diagram=other_section_name)

        # first check that default section only contains placeholder
        result = model_card.select(
            "Model description/Training Procedure/Model Plot"
        ).format()
        assert result == CONTENT_PLACEHOLDER

        # now check that the actual model diagram is in the other section
        result = model_card.select(other_section_name).format()
        assert result.startswith("<style>#sk-")
        assert "<style>" in result
        assert result.endswith(
            "<pre>LinearRegression()</pre></div></div></div></div></div>"
        )

    def test_other_section(self, model_card):
        model_card.add_model_plot(section="Other section")
        result = model_card.select("Other section").content
        assert result.startswith("<style>#sk-")
        assert "<style>" in result
        assert result.endswith(
            "<pre>LinearRegression()</pre></div></div></div></div></div>"
        )

    def test_with_description(self, model_card):
        model_card.add_model_plot(description="Awesome diagram below")
        result = model_card.select(
            "Model description/Training Procedure/Model Plot"
        ).format()
        assert result.startswith("Awesome diagram below\n\n<style>#sk-")

    @pytest.mark.parametrize("template", CUSTOM_TEMPLATES)
    def test_custom_template_no_section_uses_default(self, template):
        model = fit_model()
        model_card = Card(model, template=template)
        model_card.add_model_plot()
        result = model_card.select(
            "Model description/Training Procedure/Model Plot"
        ).format()

        # don't compare whole text, as it's quite long and non-deterministic
        assert result.startswith("<style>#sk-")
        assert "<style>" in result
        assert result.endswith(
            "<pre>LinearRegression()</pre></div></div></div></div></div>"
        )

    @pytest.mark.parametrize("template", CUSTOM_TEMPLATES)
    def test_custom_template_init_str_works(self, template):
        model = fit_model()
        section_name = "Here is the model diagram"
        model_card = Card(model, template=template, model_diagram=section_name)

        result = model_card.select(section_name).format()
        assert result.startswith("<style>#sk-")
        assert "<style>" in result
        assert result.endswith(
            "<pre>LinearRegression()</pre></div></div></div></div></div>"
        )

    def test_default_template_and_model_diagram_true(self, model_card):
        # setting model_diagram=True should not change anything vs auto with the
        # default template
        model = fit_model()
        model_card = Card(model, model_diagram=True)
        result = model_card.select(
            "Model description/Training Procedure/Model Plot"
        ).format()
        # don't compare whole text, as it's quite long and non-deterministic
        assert result.startswith("<style>#sk-")
        assert "<style>" in result
        assert result.endswith(
            "<pre>LinearRegression()</pre></div></div></div></div></div>"
        )

    @pytest.mark.parametrize("template", CUSTOM_TEMPLATES)
    def test_custom_template_and_model_diagram_true_uses_default(
        self, model_card, template
    ):
        # when using a custom template and not indicating a section, use the
        # default section
        model = fit_model()
        Card(model, template=template, model_diagram=True)
        result = model_card.select(
            "Model description/Training Procedure/Model Plot"
        ).format()
        # don't compare whole text, as it's quite long and non-deterministic
        assert result.startswith("<style>#sk-")
        assert "<style>" in result
        assert result.endswith(
            "<pre>LinearRegression()</pre></div></div></div></div></div>"
        )

    def test_add_twice(self, model_card):
        # it's possible to add the section twice, even if it doesn't make a lot
        # of sense
        text1 = model_card.select(
            "Model description/Training Procedure/Model Plot"
        ).content
        model_card.add_model_plot(section="Other section")
        text2 = model_card.select("Other section").content

        # both are identical, except for numbers like "#sk-container-id-123",
        # thus compare everything but the numbers
        assert re.split(r"\d+", text1) == re.split(r"\d+", text2)


def _strip_multiple_chars(text, char):
    # utility function needed to compare tables across systems
    # _strip_multiple_chars("hi    there") == "hi there"
    # _strip_multiple_chars("|---|--|", "-") == "|-|-|"
    while char + char in text:
        text = text.replace(char + char, char)
    return text


class TestAddHyperparams:
    """Tests for the model hyperparameters"""

    @pytest.fixture
    def expected(self):
        lines = [
            "<details>",
            "<summary> Click to expand </summary>",
            "",
            "| Hyperparameter   | Value       |",
            "|------------------|-------------|",
            "| copy_X           | True        |",
            "| fit_intercept    | True        |",
            "| n_jobs           |             |",
            "| normalize        | deprecated  |",
            "| positive         | False       |",
            "",
            "</details>",
        ]
        # TODO: After dropping sklearn < 1.2, when the "normalize" parameter is
        # removed, remove it from the table above and remove the code below.
        major, minor, *_ = sklearn.__version__.split(".")
        major, minor = int(major), int(minor)
        if (major >= 1) and (minor >= 2):
            del lines[8]

        table = "\n".join(lines)
        # remove multiple whitespaces and dashes, as they're not important and may
        # differ depending on OS
        table = _strip_multiple_chars(table, " ")
        table = _strip_multiple_chars(table, "-")
        return table

    def test_default(self, model_card, expected):
        result = model_card.select(
            "Model description/Training Procedure/Hyperparameters"
        ).format()

        # remove multiple whitespaces and dashes, as they're not important and may
        # differ depending on OS
        result = _strip_multiple_chars(result, " ")
        result = _strip_multiple_chars(result, "-")
        assert result == expected

    def test_other_section(self, model_card, expected):
        model_card.add_hyperparams(section="Other section")
        result = model_card.select("Other section").format()

        # remove multiple whitespaces and dashes, as they're not important and may
        # differ depending on OS
        result = _strip_multiple_chars(result, " ")
        result = _strip_multiple_chars(result, "-")
        assert result == expected

    def test_with_description(self, model_card, expected):
        model_card.add_hyperparams(description="Awesome hyperparams")
        result = model_card.select(
            "Model description/Training Procedure/Hyperparameters"
        ).format()
        assert result.startswith("Awesome hyperparams")

    @pytest.mark.parametrize("template", CUSTOM_TEMPLATES)
    def test_custom_template_no_section_uses_default(self, template, expected):
        model = fit_model()
        model_card = Card(model, template=template)
        model_card.add_hyperparams()
        result = model_card.select(
            "Model description/Training Procedure/Hyperparameters"
        ).format()

        # remove multiple whitespaces and dashes, as they're not important and may
        # differ depending on OS
        result = _strip_multiple_chars(result, " ")
        result = _strip_multiple_chars(result, "-")
        assert result == expected

    def test_add_twice(self, model_card):
        # it's possible to add the section twice, even if it doesn't make a lot
        # of sense
        text1 = model_card.select(
            "Model description/Training Procedure/Hyperparameters"
        ).format()
        model_card.add_hyperparams(section="Other section")
        text2 = model_card.select("Other section").format()

        assert text1 == text2

    def test_hyperparameter_table_with_line_break(self):
        # Hyperparameters can contain values with line breaks, "\n", in them. In
        # that case, the markdown table is broken. Check that the hyperparameter
        # table we create properly replaces the "\n" with "<br />".
        class EstimatorWithLbInParams:
            def get_params(self, deep=False):
                return {"fit_intercept": True, "n_jobs": "line\nwith\nbreak"}

        model_card = Card(EstimatorWithLbInParams())
        section_name = "Model description/Training Procedure/Hyperparameters"
        text_hyperparams = model_card.select(section_name).format()

        # remove multiple whitespaces, as they're not important
        text_cleaned = _strip_multiple_chars(text_hyperparams, " ")
        assert "| n_jobs | line<br />with<br />break |" in text_cleaned


class TestAddMetrics:
    """Tests for adding metrics"""

    def test_default(self, model_card):
        # by default, don't add a table, as there are no metrics
        result = model_card.select("Model description/Evaluation Results").format()
        expected = CONTENT_PLACEHOLDER
        assert result == expected

    def test_empty_metrics_table(self, model_card):
        model_card.add_metrics()
        result = model_card.select("Model description/Evaluation Results").format()
        expected = (
            "| Metric   | Value   |\n"  # fmt: skip
            "|----------|---------|"
        )
        assert result == expected

    def test_multiple_metrics(self, model_card):
        model_card.add_metrics(**{"acc": "0.1"})  # str
        model_card.add_metrics(
            f1=0.1,  # float
            awesomeness=123,  # int
        )
        result = model_card.select("Model description/Evaluation Results").format()
        expected = (
            "| Metric      |   Value |\n"
            "|-------------|---------|\n"
            "| acc         |     0.1 |\n"
            "| f1          |     0.1 |\n"
            "| awesomeness |   123   |"
        )
        assert result == expected

    def test_other_section(self, model_card):
        model_card.add_metrics(accuracy=0.9, section="Other section")
        result = model_card.select("Other section").format()
        # fmt: off
        expected = (
            "| Metric   |   Value |\n"
            "|----------|---------|\n"
            "| accuracy |     0.9 |"
        )
        # fmt: on
        assert result == expected

    def test_with_description(self, model_card):
        model_card.add_metrics(accuracy=0.9, description="Awesome metrics")
        result = model_card.select("Model description/Evaluation Results").format()
        assert result.startswith("Awesome metrics\n\n| Metric ")

    @pytest.mark.parametrize("template", CUSTOM_TEMPLATES)
    def test_custom_template_no_section_raises(self, template):
        model = fit_model()
        model_card = Card(model, template=template)
        model_card.add_metrics(accuracy=0.9)

        result = model_card.select("Model description/Evaluation Results").format()
        # fmt: off
        expected = (
            "| Metric   |   Value |\n"
            "|----------|---------|\n"
            "| accuracy |     0.9 |"
        )
        # fmt: on
        assert result == expected

    def test_add_twice(self, model_card):
        # it's possible to add the section twice, even if it doesn't make a lot
        # of sense
        model_card.add_metrics(accuracy=0.9)
        text1 = model_card.select("Model description/Evaluation Results").format()
        model_card.add_metrics(section="Other section")
        text2 = model_card.select("Other section").format()
        assert text1 == text2


class TestAddPermutationImportance:
    @pytest.fixture
    def importances(self, iris_estimator, iris_data):
        X, y = iris_data
        result = permutation_importance(
            iris_estimator, X, y, n_repeats=10, random_state=42, n_jobs=2
        )
        return result

    def test_permutation_importances(
        self, iris_data, importances, model_card, destination_path
    ):
        X, _ = iris_data
        model_card.add_permutation_importances(
            importances,
            columns=X.columns,
            plot_file=Path(destination_path) / "importance.png",
            plot_name="Permutation Importance",
        )
        temp_path = Path(destination_path) / "importance.png"
        section = model_card.select("Permutation Importance")
        expected = f"![Permutation Importance]({temp_path})"
        assert section.format() == expected

    def test_multiple_permutation_importances(
        self, iris_data, iris_estimator, importances, model_card, destination_path
    ):
        X, y = iris_data
        model_card.add_permutation_importances(
            importances, X.columns, plot_file=Path(destination_path) / "importance.png"
        )

        f1 = make_scorer(f1_score, average="micro")
        importances_f1 = permutation_importance(
            iris_estimator, X, y, scoring=f1, n_repeats=10, random_state=42, n_jobs=2
        )
        model_card.add_permutation_importances(
            importances_f1,
            columns=X.columns,
            plot_file=Path(destination_path) / "f1_importance.png",
            plot_name="Permutation Importance on f1",
        )

        # check for default one
        temp_path = Path(destination_path) / "importance.png"
        section = model_card.select("Permutation Importances")
        expected = f"![Permutation Importances]({temp_path})"
        assert section.format() == expected

        # check for F1
        temp_path_f1 = Path(destination_path) / "f1_importance.png"
        section = model_card.select("Permutation Importance on f1")
        expected = f"![Permutation Importance on f1]({temp_path_f1})"
        assert section.format() == expected

    def test_duplicate_permutation_importances(
        self, iris_data, importances, model_card, destination_path
    ):
        X, _ = iris_data
        plot_path = os.path.join(destination_path, "importance.png")
        model_card.add_permutation_importances(
            importances, X.columns, plot_file=plot_path
        )
        with pytest.raises(
            ValueError,
            match=(
                "already exists. Set `overwrite` to `True` or pass a"
                " different filename for the plot."
            ),
        ):
            model_card.add_permutation_importances(
                importances,
                columns=X.columns,
                plot_file=plot_path,
                plot_name="Permutation Importance on f1",
            )

    def test_duplicate_permutation_importances_overwrite(
        self, iris_data, importances, model_card, destination_path
    ):
        X, _ = iris_data
        plot_path = os.path.join(destination_path, "importance.png")
        model_card.add_permutation_importances(
            importances, X.columns, plot_file=plot_path
        )

        model_card.add_permutation_importances(
            importances,
            columns=X.columns,
            plot_file=plot_path,
            plot_name="Permutation Importance on f1",
            overwrite=True,
        )
        section = model_card.select("Permutation Importance on f1")
        expected = f"![Permutation Importance on f1]({plot_path})"
        assert section.format() == expected

    def test_permutation_importances_with_description(
        self, iris_data, importances, model_card, destination_path
    ):
        X, _ = iris_data
        model_card.add_permutation_importances(
            importances,
            columns=X.columns,
            plot_file=Path(destination_path) / "importance.png",
            description="Very important",
        )
        temp_path = Path(destination_path) / "importance.png"
        section = model_card.select("Permutation Importances")
        expected = f"Very important\n\n![Permutation Importances]({temp_path})"
        assert section.format() == expected


class TestAddGetStartedCode:
    """Tests for getting started code"""

    @pytest.fixture
    def metadata(self):
        # dummy ModelCardData using pickle
        class Metadata:
            def to_dict(self):
                return {
                    "model_file": "my-model.pickle",
                    "sklearn": {
                        "model_format": "pickle",
                    },
                }

        return Metadata()

    @pytest.fixture
    def model_card(self, metadata):
        model = fit_model()
        card = Card(model, metadata=metadata)
        return card

    @pytest.fixture
    def metadata_skops(self):
        # dummy ModelCardData using skops
        class Metadata:
            def to_dict(self):
                return {
                    "model_file": "my-model.skops",
                    "sklearn": {
                        "model_format": "skops",
                    },
                }

        return Metadata()

    @pytest.fixture
    def model_card_skops(self, metadata_skops):
        model = fit_model()
        card = Card(model, metadata=metadata_skops)
        return card

    @pytest.fixture
    def text_pickle(self):
        return (
            "```python\n"
            "import json\n"
            "import pandas as pd\n"
            "import joblib\n"
            'model = joblib.load("my-model.pickle")\n'
            'with open("config.json") as f:\n'
            "    config = json.load(f)\n"
            'model.predict(pd.DataFrame.from_dict(config["sklearn"]["example_input"]))\n'
            "```"
        )

    @pytest.fixture
    def text_skops(self):
        return (
            "```python\n"
            "import json\n"
            "import pandas as pd\n"
            "import skops.io as sio\n"
            'model = sio.load("my-model.skops")\n'
            'with open("config.json") as f:\n'
            "    config = json.load(f)\n"
            'model.predict(pd.DataFrame.from_dict(config["sklearn"]["example_input"]))\n'
            "```"
        )

    def test_default_pickle(self, model_card, text_pickle):
        # by default, don't add a table, as there are no metrics
        result = model_card.select("How to Get Started with the Model").format()
        assert result == text_pickle

    def test_default_skops(self, model_card_skops, text_skops):
        # by default, don't add a table, as there are no metrics
        result = model_card_skops.select("How to Get Started with the Model").format()
        assert result == text_skops

    def test_no_metadata_file_name(self):
        model = fit_model()
        card = Card(model, metadata=None)
        card.add_get_started_code()  # does not raise

    def test_no_metadata_file_format(self):
        class Metadata:
            def to_dict(self):
                return {
                    "model_file": "my-model.skops",
                    # missing file format entry
                }

        model = fit_model()
        card = Card(model, metadata=Metadata())
        card.add_get_started_code()  # does not raise

    def test_other_section(self, model_card, text_pickle):
        model_card.add_get_started_code(section="Other section")
        result = model_card.select("Other section").format()
        assert result == text_pickle

    def test_use_description(self, model_card):
        model_card.add_get_started_code(description="Awesome code")
        result = model_card.select("How to Get Started with the Model").format()
        assert result.startswith("Awesome code")

    def test_other_filename(self, model_card, text_pickle):
        model_card.add_get_started_code(file_name="foobar.pkl")
        text = text_pickle.replace("my-model.pickle", "foobar.pkl")
        result = model_card.select("How to Get Started with the Model").format()
        assert result == text

    def test_explicitly_set_other_model_format(self, model_card, text_skops):
        model_card.add_get_started_code(model_format="skops")
        result = model_card.select("How to Get Started with the Model").format()
        # file name is still "my-model.pickle", only the loading code changes
        text = text_skops.replace(".skops", ".pickle")
        assert result == text

    def test_invalid_model_format_passed(self, model_card):
        # json is not a valid model format
        msg = "Invalid model format 'json', should be one of 'pickle' or 'skops'"
        with pytest.raises(ValueError, match=msg):
            model_card.add_get_started_code(model_format="json")

    def test_invalid_model_format_passed_via_metadata(self):
        # metadata contains invalid model format json
        class Metadata:
            def to_dict(self):
                return {
                    "model_file": "my-model.skops",
                    "sklearn": {
                        "model_format": "json",
                    },
                }

        model = fit_model()

        msg = "Invalid model format 'json', should be one of 'pickle' or 'skops'"
        with pytest.raises(ValueError, match=msg):
            Card(model, metadata=Metadata())

    @pytest.mark.parametrize("template", CUSTOM_TEMPLATES)
    def test_custom_template_no_section_uses_default(self, template, text_pickle):
        model = fit_model()

        class Metadata:
            def to_dict(self):
                return {
                    "model_file": "my-model.pickle",
                    "sklearn": {
                        "model_format": "pickle",
                    },
                }

        model_card = Card(model, metadata=Metadata(), template=template)
        model_card.add_get_started_code()
        result = model_card.select("How to Get Started with the Model").format()
        assert result == text_pickle

    def test_add_twice(self, model_card):
        # it's possible to add the section twice, even if it doesn't make a lot
        # of sense
        text1 = model_card.select("How to Get Started with the Model").format()
        model_card.add_get_started_code(section="Other section")
        text2 = model_card.select("Other section").format()
        assert text1 == text2


class TestRender:
    def test_render(self, model_card, destination_path):
        file_name = destination_path / "README.md"
        model_card.save(file_name)
        with open(file_name, "r", encoding="utf-8") as f:
            loaded = f.read()

        rendered = model_card.render()
        assert loaded == rendered

    def test_render_with_metadata(self, model_card):
        model_card.metadata.foo = "something"
        model_card.metadata.bar = "something else"
        rendered = model_card.render()
        expected = textwrap.dedent(
            """
            ---
            foo: something
            bar: something else
            ---
            """
        ).strip()
        assert rendered.startswith(expected)


class TestSelect:
    """Selecting sections from the model card"""

    def test_select_existing_section(self, model_card):
        section = model_card.select("Model description")
        assert section.title == "Model description"

    def test_select_existing_subsection(self, model_card):
        section = model_card.select("Model description/Training Procedure")
        assert section.title == "Training Procedure"

        section = model_card.select("Model description").select("Training Procedure")
        assert section.title == "Training Procedure"

    def test_select_existing_subsubsection(self, model_card):
        section = model_card.select(
            "Model description/Training Procedure/Hyperparameters"
        )
        assert section.title == "Hyperparameters"

        section = (
            model_card.select("Model description")
            .select("Training Procedure")
            .select("Hyperparameters")
        )
        assert section.title == "Hyperparameters"

    def test_select_non_existing_section_raises(self, model_card):
        with pytest.raises(KeyError):
            model_card.select("non-existing section")

    def test_select_non_existing_subsection_raises(self, model_card):
        with pytest.raises(KeyError):
            model_card.select("Model description/non-existing subsection")

        with pytest.raises(KeyError):
            model_card.select("Model description").select("non-existing subsection")

    def test_select_non_existing_subsubsection_raises(self, model_card):
        msg = "non-existing sub-subsection"

        with pytest.raises(KeyError, match=msg):
            model_card.select(
                "Model description/Training Procedure/non-existing sub-subsection"
            )

        with pytest.raises(KeyError, match=msg):
            (
                model_card.select("Model description")
                .select("Training Procedure")
                .select("non-existing sub-subsection")
            )

    def test_select_non_existing_section_and_subsection_raises(self, model_card):
        msg = "non-existing section"

        with pytest.raises(KeyError, match=msg):
            model_card.select("non-existing section/non-existing subsection")

        with pytest.raises(KeyError, match=msg):
            model_card.select("non-existing section").select("non-existing subsection")

    def test_select_empty_key_raises(self, model_card):
        msg = r"Section name cannot be empty but got ''"
        with pytest.raises(KeyError, match=msg):
            model_card.select("")

    def test_select_empty_key_subsection_raises(self, model_card):
        msg = r"Section name cannot be empty but got 'Model description/'"
        with pytest.raises(KeyError, match=msg):
            model_card.select("Model description/")

        msg = r"Section name cannot be empty but got ''"
        with pytest.raises(KeyError, match=msg):
            model_card.select("Model description").select("")

    def test_default_skops_sections_present(self, model_card):
        # model_card (which is prefilled) contains all default sections
        for key in SKOPS_TEMPLATE:
            model_card.select(key)

    def test_default_skops_sections_empty_card(self, model_card):
        # Without prefilled template, the card should not contain the default sections

        # empty card does not contain any sections, so trying to select them
        # should raise a KeyError
        model = fit_model()
        card_empty = Card(model, model_diagram=False, template=None)
        for key in SKOPS_TEMPLATE:
            with pytest.raises(KeyError):
                card_empty.select(key)

    def test_invalid_template_name_raises(self):
        msg = "Unknown template 'does-not-exist', template must be one of the following"
        with pytest.raises(ValueError, match=msg):
            Card(model=None, template="does-not-exist")


class TestAdd:
    """Adding sections and subsections"""

    def test_add_new_section(self, model_card):
        model_card = model_card.add(**{"A new section": "sklearn FTW"})
        section = model_card.select("A new section")
        assert section.title == "A new section"
        assert section.content == "sklearn FTW"

    def test_add_new_subsection(self, model_card):
        model_card = model_card.add(
            **{"Model description/A new section": "sklearn FTW"}
        )
        section = model_card.select("Model description/A new section")
        assert section.title == "A new section"
        assert section.content == "sklearn FTW"

        # make sure that the new subsection is the last subsection
        subsections = model_card._data["Model description"].subsections
        assert len(subsections) > 1  # exclude trivial case of only one subsection

        last_subsection = list(subsections.values())[-1]
        assert last_subsection is section

    def test_add_new_section_and_subsection(self, model_card):
        model_card = model_card.add(**{"A new section/A new subsection": "sklearn FTW"})

        section = model_card.select("A new section")
        assert section.title == "A new section"
        assert section.content == ""

        subsection = model_card.select("A new section/A new subsection")
        assert subsection.title == "A new subsection"
        assert subsection.content == "sklearn FTW"

    def test_add_new_section_with_slash_in_name(self, model_card):
        model_card = model_card.add(**{"A new\\/section": "sklearn FTW"})
        section = model_card.select("A new\\/section")
        assert section.title == "A new/section"
        assert section.content == "sklearn FTW"

    def test_add_new_subsection_with_slash_in_name(self, model_card):
        model_card = model_card.add(
            **{"Model description/A new\\/section": "sklearn FTW"}
        )
        section = model_card.select("Model description/A new\\/section")
        assert section.title == "A new/section"
        assert section.content == "sklearn FTW"

    def test_add_content_to_existing_section(self, model_card):
        # Add content (not new sections) to an existing section. Make sure that
        # existing subsections are not affected by this
        section = model_card.select("Model description")
        num_subsection_before = len(section.subsections)
        assert num_subsection_before > 0  # exclude trivial case of empty sections

        # add content to "Model description" section
        model_card = model_card.add(**{"Model description": "sklearn FTW"})
        section = model_card.select("Model description")
        num_subsection_after = len(section.subsections)

        assert num_subsection_before == num_subsection_after
        assert section.content == "sklearn FTW"

    def test_add_plain_section_works(self, model_card):
        # It is allowed to add a *Section object, but it's not documented and
        # users should normally not use that feature
        section = Section("title may differ from section name", "some content")
        model_card.add(
            a_string="normal string",
            a_section=section,
        )
        assert model_card.select("a_section") == section

    def test_add_section_preserves_subsections(self, model_card):
        # As explained in the previous test, users can theoretically add section
        # instances. If they override an existing section with a new section,
        # the subsections of the existing section should be preserved.

        # first let's add a section and a subsection
        model_card.add(**{"new section": "hello", "new section/subsection": "world"})
        assert model_card.select("new section").format() == "hello"
        assert model_card.select("new section/subsection").format() == "world"

        # now let's override the section, the subsection should be preserved
        new_section = Section("new section", "bonjour")
        model_card.add(**{"new section": new_section})
        assert model_card.select("new section").format() == "bonjour"
        assert model_card.select("new section/subsection").format() == "world"

    def test_add_section_with_identical_subsection_preserves_subsections(
        self, model_card
    ):
        # As explained in the previous tests, users can theoretically add
        # section instances. If they override an existing section with a new
        # section, the subsections of the existing section should be preserved.
        # If the new section they add has its own subsections, and these
        # subsections are identical to the old subsections, that should be fine.

        # first let's add a section and a subsection
        model_card.add(**{"new section": "hello", "new section/subsection": "world"})

        # now let's override the section using the same subsections
        old_subsection = model_card.select("new section").subsections
        new_section = Section("new section", "bonjour", subsections=old_subsection)
        model_card.add(**{"new section": new_section})
        assert model_card.select("new section").format() == "bonjour"
        assert model_card.select("new section/subsection").format() == "world"

    def test_add_section_with_different_subsection_raises(self, model_card):
        # This is the same as the previous test, but now the section used to
        # override the previous section has different subsections. Now we don't
        # know what to do and should raise. This is okay because normally, a
        # user shouldn't add section instances anyway.

        # first let's add a section and a subsection
        model_card.add(**{"new section": "hello", "new section/subsection": "world"})

        # now let's override the section using different subsections
        new_subsection = {"new subsection": Section("subsection", "mars")}
        new_section = Section("new section", "bonjour", subsections=new_subsection)

        match = (
            "Trying to override section 'new section' but found conflicting subsections"
        )
        with pytest.raises(ValueError, match=match):
            model_card.add(**{"new section": new_section})


class TestDelete:
    """Deleting sections and subsections"""

    def test_delete_section(self, model_card):
        model_card.select("Model description")
        model_card.delete("Model description")
        with pytest.raises(KeyError):
            model_card.select("Model description")

    def test_delete_subsection(self, model_card):
        model_card.select("Model description/Training Procedure")
        model_card.delete("Model description/Training Procedure")
        with pytest.raises(KeyError):
            model_card.select("Model description/Training Procedure")
        # parent section still exists
        model_card.delete("Model description")

    def test_delete_subsubsection(self, model_card):
        model_card.select("Model description/Training Procedure/Hyperparameters")
        model_card.delete("Model description/Training Procedure/Hyperparameters")
        with pytest.raises(KeyError):
            model_card.select("Model description/Training Procedure/Hyperparameters")
        # parent section still exists
        model_card.delete("Model description/Training Procedure")

    def test_delete_section_with_slash_in_name(self, model_card):
        model_card.add(**{"A new\\/section": "some content"})
        model_card.select("A new\\/section")
        model_card.delete("A new\\/section")
        with pytest.raises(KeyError):
            model_card.select("A new\\/section")

    def test_delete_non_existing_section_raises(self, model_card):
        with pytest.raises(KeyError):
            model_card.delete("non-existing section")

    def test_delete_non_existing_subsection_raises(self, model_card):
        with pytest.raises(KeyError):
            model_card.delete("Model description/non-existing subsection")

        with pytest.raises(KeyError):
            model_card.delete(["Model description", "non-existing subsection"])

    def test_delete_non_existing_subsubsection_raises(self, model_card):
        with pytest.raises(KeyError):
            model_card.delete(
                "Model description/Training Procedure/non-existing sub-subsection"
            )

        with pytest.raises(KeyError):
            model_card.delete(
                [
                    "Model description",
                    "Training Procedure",
                    "non-existing sub-subsection",
                ]
            )

    def test_delete_non_existing_section_and_subsection_raises(self, model_card):
        with pytest.raises(KeyError):
            model_card.delete(["non-existing section", "non-existing subsection"])

    def test_delete_empty_key_raises(self, model_card):
        msg = r"Section name cannot be empty but got ''"
        with pytest.raises(KeyError, match=msg):
            model_card.delete("")

        msg = r"Section name cannot be empty but got '\[\]'"
        with pytest.raises(KeyError, match=msg):
            model_card.delete([])

    def test_delete_empty_key_subsection_raises(self, model_card):
        msg = r"Section name cannot be empty but got 'Model description/'"
        with pytest.raises(KeyError, match=msg):
            model_card.delete("Model description/")

        msg = r"Section name cannot be empty but got '\['Model description', ''\]'"
        with pytest.raises(KeyError, match=msg):
            model_card.delete(["Model description", ""])


class TestAddPlot:
    def test_add_plot(self, destination_path, model_card):
        import matplotlib.pyplot as plt

        plt.plot([4, 5, 6, 7])
        plt.savefig(Path(destination_path) / "fig1.png")
        model_card = model_card.add_plot(fig1="fig1.png")
        plot_content = model_card.select("fig1").format()
        assert plot_content == "![fig1](fig1.png)"

    def test_add_plot_to_existing_section(self, destination_path, model_card):
        import matplotlib.pyplot as plt

        plt.plot([4, 5, 6, 7])
        plt.savefig(Path(destination_path) / "fig1.png")
        model_card = model_card.add_plot(**{"Model description/Figure 1": "fig1.png"})
        plot_content = model_card.select("Model description/Figure 1").format()
        assert plot_content == "![Figure 1](fig1.png)"

    def test_add_plot_with_description(self, destination_path, model_card):
        import matplotlib.pyplot as plt

        plt.plot([4, 5, 6, 7])
        plt.savefig(Path(destination_path) / "fig1.png")
        model_card = model_card.add_plot(description="My fancy plot", fig1="fig1.png")
        plot_content = model_card.select("fig1").format()
        assert plot_content == "My fancy plot\n\n![fig1](fig1.png)"

    def test_add_plot_with_alt_text(self, destination_path, model_card):
        import matplotlib.pyplot as plt

        plt.plot([4, 5, 6, 7])
        plt.savefig(Path(destination_path) / "fig1.png")
        model_card = model_card.add_plot(alt_text="the figure", fig1="fig1.png")
        plot_content = model_card.select("fig1").format()
        assert plot_content == "![the figure](fig1.png)"


class TestMetadata:
    def test_adding_metadata(self, model_card):
        # test if the metadata is added to the card
        model_card.metadata.tags = "dummy"
        metadata = list(model_card._generate_metadata(model_card.metadata))
        assert len(metadata) == 1
        assert metadata[0] == "metadata.tags=dummy,"

    def test_metadata_from_config_tabular_data(
        self, pkl_model_card_metadata_from_config, destination_path
    ):
        # test if widget data is correctly set in the README
        metadata = metadata_load(local_path=Path(destination_path) / "README.md")
        assert "widget" in metadata

        expected_data = {
            "structuredData": {
                "petal length (cm)": [1.4, 1.4, 1.3],
                "petal width (cm)": [0.2, 0.2, 0.2],
                "sepal length (cm)": [5.1, 4.9, 4.7],
                "sepal width (cm)": [3.5, 3.0, 3.2],
            }
        }
        assert metadata["widget"] == expected_data

        for tag in ["sklearn", "skops", "tabular-classification"]:
            assert tag in metadata["tags"]

    def test_metadata_model_format_pkl(
        self, pkl_model_card_metadata_from_config, destination_path
    ):
        metadata = metadata_load(local_path=Path(destination_path) / "README.md")
        assert metadata["model_format"] == "pickle"

    def test_metadata_model_format_skops(
        self, skops_model_card_metadata_from_config, destination_path
    ):
        metadata = metadata_load(local_path=Path(destination_path) / "README.md")
        assert metadata["model_format"] == "skops"

    def test_metadata_tags_without_sklearn_intelex_tag(
        self, destination_path, iris_data, iris_pkl_file
    ):
        # by default, intelex is not being used
        X, _ = iris_data
        hub_utils.init(
            model=iris_pkl_file,
            requirements=[],
            dst=destination_path,
            task="tabular-classification",
            data=X,
        )

        metadata = metadata_from_config(destination_path)
        assert "scikit-learn-intelex" not in metadata.tags

    def test_metadata_tags_with_sklearn_intelex_tag(
        self, destination_path, iris_data, iris_pkl_file
    ):
        X, _ = iris_data
        hub_utils.init(
            model=iris_pkl_file,
            requirements=[],
            dst=destination_path,
            task="tabular-classification",
            data=X,
            use_intelex=True,
        )

        metadata = metadata_from_config(destination_path)
        assert "scikit-learn-intelex" in metadata.tags


@pytest.mark.xfail(reason="dynamic adjustment when model changes not implemented yet")
class TestModelDynamicUpdate:
    def test_model_related_sections_updated_dynamically_skops_template(
        self, model_card
    ):
        # Change the model to be a KNN classifier and check that the sections
        # related to the model, the plot and hyperparams, are updated correctly.
        # But first, as a sanity check, ensure that before the change, there is
        # no reference to KNN.
        model_plot_before = model_card.select(
            "Model description/Training Procedure/Model Plot"
        )
        assert "KNeighborsClassifier" not in model_plot_before.content

        hyperparams_before = model_card.select(
            "Model description/Training Procedure/Hyperparameters"
        )
        assert "n_neighbors" not in hyperparams_before.content.format()

        # change model to KNN
        model_after = KNeighborsClassifier()
        model_card.model = model_after

        model_plot_after = model_card.select(
            "Model description/Training Procedure/Model Plot"
        ).content
        assert "KNeighborsClassifier" in model_plot_after

        hyperparams_after = model_card.select(
            "Model description/Training Procedure/Hyperparameters"
        )
        assert "n_neighbors" in hyperparams_after.content.format()

    def test_model_related_sections_updated_dynamically_custom_template(
        self, model_card
    ):
        # same as previous test but using a custom template
        template = {"My model plot": ""}
        model = fit_model()
        model_card = Card(model, template=template)

        # add model plot and hyperparams -- section must be passed but it
        # doesn't need to already exist in the custom template
        model_card.add_model_plot(section="My model plot")
        model_card.add_hyperparams(section="My hyperparams")

        model_plot_before = model_card.select("My model plot")
        assert "KNeighborsClassifier" not in model_plot_before.content

        hyperparams_before = model_card.select("My hyperparams")
        assert "n_neighbors" not in hyperparams_before.content.format()

        # change model to KNN
        model_after = KNeighborsClassifier()
        model_card.model = model_after

        model_plot_after = model_card.select("My model plot")
        assert "KNeighborsClassifier" in model_plot_after.content

        hyperparams_after = model_card.select("My hyperparams")
        assert "n_neighbors" in hyperparams_after.content.format()


class TestCardRepr:
    """Test __str__ and __repr__ methods of Card, which are identical for now"""

    @pytest.fixture
    def card(self):
        model = LinearRegression(fit_intercept=False)
        card = Card(model=model)
        card.add(Figures="")
        card.add(
            **{
                "Model Description": "A description",
                "Model Card Authors": "Jane Doe",
            }
        )
        card.add_plot(
            **{
                "Figures/ROC": "ROC.png",
                "Figures/Confusion matrix": "confusion_matrix.jpg",
            }
        )
        card.add_table(**{"Search Results": {"split": [1, 2, 3], "score": [4, 5, 6]}})
        return card

    @pytest.fixture
    def expected_lines(self):
        card_repr = """
        Card(
          model=LinearRegression(fit_intercept=False),
          Model description/Training Procedure/Hyperparameters=TableSection(4x2),
          Model description/Training Procedure/...</pre></div></div></div></div></div>,
          Model Card Authors=Jane Doe,
          Figures/ROC=PlotSection(ROC.png),
          Figures/Confusion matrix=PlotSection(confusion_matrix.jpg),
          Model Description=A description,
          Search Results=TableSection(3x2),
        )
        """
        expected = textwrap.dedent(card_repr).strip()
        lines = expected.split("\n")
        return lines

    @pytest.mark.parametrize("meth", [repr, str])
    def test_card_repr(self, card: Card, meth, expected_lines):
        result = meth(card)
        expected = "\n".join(expected_lines)
        assert result == expected

    @pytest.mark.parametrize("meth", [repr, str])
    def test_card_repr_empty_card(self, meth):
        """Without prefilled template, the repr should be empty"""
        model = fit_model()
        card = Card(model, model_diagram=False, template=None)
        result = meth(card)
        expected = textwrap.dedent(
            """
        Card(
          model=LinearRegression(),
        )
        """
        ).strip()
        assert result == expected

    @pytest.mark.parametrize("meth", [repr, str])
    def test_very_long_lines_are_shortened(self, card: Card, meth, expected_lines):
        card.add(my_section="very long line " * 100)

        # expected results contain 1 line at the very end
        extra_line = (
            "  my_section=very long line very long l... "
            "line very long line very long line ,"
        )
        expected_lines.insert(-1, extra_line)
        expected = "\n".join(expected_lines)

        result = meth(card)
        assert result == expected

    @pytest.mark.parametrize("meth", [repr, str])
    def test_without_model_attribute(self, card: Card, meth, expected_lines):
        del card.model

        # remove line 1 from expected results, which corresponds to the model
        del expected_lines[1]
        expected = "\n".join(expected_lines)

        result = meth(card)
        assert result == expected

    @pytest.mark.parametrize("meth", [repr, str])
    def test_with_metadata(self, card: Card, meth, expected_lines):
        metadata = ModelCardData(
            language="fr",
            license="bsd",
            library_name="sklearn",
            tags=["sklearn", "tabular-classification"],
            foo={"bar": 123},
            widget={"something": "very-long"},
        )
        card.metadata = metadata

        # metadata comes after model line, i.e. position 2
        extra_lines = [
            "  metadata.language=fr,",
            "  metadata.license=bsd,",
            "  metadata.library_name=sklearn,",
            "  metadata.tags=['sklearn', 'tabular-classification'],",
            "  metadata.foo={'bar': 123},",
            "  metadata.widget={...},",
        ]
        expected = "\n".join(expected_lines[:2] + extra_lines + expected_lines[2:])

        result = meth(card)
        assert result == expected


class TestCardModelAttributeIsPath:
    def path_to_card(self, path):
        card = Card(model=path, trusted=True)
        return card

    @pytest.mark.parametrize("meth", [repr, str])
    @pytest.mark.parametrize("suffix", [".pkl", ".skops"])
    def test_model_card_repr(self, meth, suffix):
        # Test that if the model is changed, Card takes this into account, if
        # the model argument is a path to a model file. First, we test that if
        # the model path changes, the Card changes. Then we test that if the
        # file on disk changes, the Card changes.
        model = LinearRegression(fit_intercept=False)
        file_handle, file_name = save_model_to_file(model, suffix)
        os.close(file_handle)
        card_from_path = self.path_to_card(file_name)

        result0 = meth(card_from_path)
        expected = "Card(\n  model=LinearRegression(fit_intercept=False),"
        assert result0.startswith(expected)

        # change file name, same card should show different result
        model = LinearRegression()
        file_handle, file_name = save_model_to_file(model, suffix)
        card_from_path.model = file_name
        result1 = meth(card_from_path)
        expected = "Card(\n  model=LinearRegression(),"
        assert result1.startswith(expected)

        # change model on disk but keep same file name, should show different
        # result
        model = LinearRegression(fit_intercept=None)
        with open(file_name, "wb") as f:
            dump_fn = pickle.dump if suffix == ".pkl" else dump
            dump_fn(model, f)
        result2 = meth(card_from_path)
        expected = "Card(\n  model=LinearRegression(fit_intercept=None),"
        assert result2.startswith(expected)

    @pytest.mark.parametrize("suffix", [".pkl", ".skops"])
    @pytest.mark.parametrize("meth", [repr, str])
    def test_load_model_exception(self, meth, suffix):
        file_handle, file_name = tempfile.mkstemp(suffix=suffix, prefix="skops-test")

        os.close(file_handle)

        with pytest.raises(Exception, match="occurred during model loading."):
            card = Card(file_name)
            meth(card)

    @pytest.mark.parametrize("meth", [repr, str])
    def test_load_model_file_not_found(self, meth):
        file_handle, file_name = tempfile.mkstemp(suffix=".pkl", prefix="skops-test")

        os.close(file_handle)
        os.remove(file_name)

        with pytest.raises(FileNotFoundError) as excinfo:
            card = Card(file_name)
            meth(card)

        assert file_name in str(excinfo.value)


class TestPlotSection:
    def test_format_path_is_str(self):
        section = PlotSection(
            title="", content="", alt_text="some title", path="path/plot.png"
        )
        expected = "![some title](path/plot.png)"
        assert section.format() == expected

    def test_format_path_is_pathlib(self):
        section = PlotSection(
            title="", content="", alt_text="some title", path=Path("path") / "plot.png"
        )
        expected = f"![some title](path{os.path.sep}plot.png)"
        assert section.format() == expected

    @pytest.mark.parametrize("meth", [str, repr])
    def test_str_and_repr(self, meth):
        section = PlotSection(
            title="", content="", alt_text="some title", path="path/plot.png"
        )
        expected = "PlotSection(path/plot.png)"
        assert meth(section) == expected

    def test_str(self):
        section = PlotSection(
            title="", content="", alt_text="some title", path="path/plot.png"
        )
        expected = "PlotSection(path/plot.png)"
        assert str(section) == expected

    @pytest.mark.parametrize("folded", [True, False])
    def test_folded(self, folded):
        section = PlotSection(
            title="",
            content="",
            alt_text="some title",
            path="path/plot.png",
            folded=folded,
        )
        output = section.format()
        if folded:
            assert "<details>" in output
        else:
            assert "<details>" not in output

    def test_add_with_description(self):
        # FIXME
        pass


class TestTableSection:
    @pytest.fixture
    def table_dict(self):
        return {"split": [1, 2, 3], "score": [4, 5, 6]}

    def test_table_is_dict(self, table_dict):
        section = TableSection(title="", content="", table=table_dict)
        expected = """|   split |   score |
|---------|---------|
|       1 |       4 |
|       2 |       5 |
|       3 |       6 |"""
        assert section.format() == expected

    def test_table_is_dataframe(self, table_dict):
        pd = pytest.importorskip("pandas")
        df = pd.DataFrame(table_dict)
        section = TableSection(title="", content="", table=df)
        expected = """|   split |   score |
|---------|---------|
|       1 |       4 |
|       2 |       5 |
|       3 |       6 |"""
        assert section.format() == expected

    @pytest.mark.parametrize("meth", [str, repr])
    def test_str_and_repr_table_is_dict(self, table_dict, meth):
        section = TableSection(title="", content="", table=table_dict)
        expected = "TableSection(3x2)"
        assert meth(section) == expected

    @pytest.mark.parametrize("meth", [str, repr])
    def test_str_and_repr_table_is_dataframe(self, table_dict, meth):
        pd = pytest.importorskip("pandas")
        df = pd.DataFrame(table_dict)
        section = TableSection(title="", content="", table=df)
        expected = "TableSection(3x2)"
        assert meth(section) == expected

    @pytest.mark.parametrize("table", [{}, "pandas"])
    def test_raise_error_empty_table(self, table):
        # If there are no columns, raise
        if table == "pandas":
            pd = pytest.importorskip("pandas")
            table = pd.DataFrame([])

        msg = "Trying to add table with no columns"
        with pytest.raises(ValueError, match=msg):
            TableSection(title="", content="", table=table)

    @pytest.mark.parametrize("table", [{"col0": []}, "pandas"])
    def test_table_with_no_rows_works(self, table):
        # If there are no rows, it's okay
        if table == "pandas":
            pd = pytest.importorskip("pandas")
            table = pd.DataFrame(data=[], columns=["col0"])

        TableSection(title="", content="", table=table).format()  # no error raised

    def test_pandas_not_installed(self, table_dict, pandas_not_installed):
        # use pandas_not_installed fixture from conftest.py to pretend that
        # pandas is not installed
        section = TableSection(title="", content="", table=table_dict)
        assert section._is_pandas_df is False

    @pytest.mark.parametrize("folded", [True, False])
    def test_folded(self, table_dict, folded):
        section = TableSection(title="", content="", table=table_dict, folded=folded)
        output = section.format()
        if folded:
            assert "<details>" in output
        else:
            assert "<details>" not in output

    def test_line_break_in_entry(self, table_dict):
        # Line breaks are not allowed inside markdown tables, so check that
        # they're removed. We test 3 conditions here:

        # 1. custom object with line breaks in repr
        # 2. string with line break in the middle
        # 3. string with line break at start, middle, and end

        # Note that for the latter, tabulate will automatically strip the line
        # breaks from the start and end.
        class LineBreakInRepr:
            """Custom object whose repr has a line break"""

            def __repr__(self) -> str:
                return "obj\nwith lb"

        table_dict["with break"] = [
            LineBreakInRepr(),
            "hi\nthere",
            """
entry with
line breaks
""",
        ]
        section = TableSection(title="", content="", table=table_dict)
        expected = """| split | score | with break |
|-|-|-|
| 1 | 4 | obj<br />with lb |
| 2 | 5 | hi<br />there |
| 3 | 6 | entry with<br />line breaks |"""

        result = section.format()
        # remove multiple whitespaces and dashes, as they're not important
        result = _strip_multiple_chars(result, " ")
        result = _strip_multiple_chars(result, "-")
        assert result == expected

    def test_add_table_with_description(self, model_card, table_dict):
        model_card.add_table(description="My fancy table", **{"The table": table_dict})
        section = model_card.select("The table")
        content = section.format()
        expected = """My fancy table

|   split |   score |
|---------|---------|
|       1 |       4 |
|       2 |       5 |
|       3 |       6 |"""
        assert content == expected


class TestCustomTemplate:
    @pytest.fixture
    def template(self):
        return {
            "My description": "An awesome model",
            "Model": "Here goes model related stuff",
            "Model/Metrics": "",
            "Foo/Bar": "Baz",
        }

    @pytest.fixture
    def card(self, template):
        model = fit_model()
        card = Card(model, template=template)
        return card

    def test_add_model_plot(self, card):
        card.add_model_plot(section="Model/Model plot")
        content = card.select("Model/Model plot").format()
        assert "LinearRegression" in content

    def test_add_hyperparams(self, card):
        card.add_hyperparams(section="Model/Hyperparams")
        content = card.select("Model/Hyperparams").format()
        assert "fit_intercept" in content

    def test_add_metrics(self, card):
        card.add_metrics(accuracy=0.1, section="Model/Metrics")
        content = card.select("Model/Metrics").format()
        assert "accuracy" in content
        assert "0.1" in content

    def test_add_get_started_code(self, card):
        card.add_get_started_code(
            section="Getting Started",
            file_name="foobar.skops",
            model_format="skops",
        )
        content = card.select("Getting Started").content
        assert "load" in content

    def test_custom_template_all_sections_present(self, template, card):
        # model_card contains all default sections
        for key in template:
            card.select(key)

        # no other top level sections as those defined in the template
        expected = ["My description", "Model", "Foo"]
        assert list(card._data.keys()) == expected


class TestRenderedCardVisibility:
    """Check that visibility flag works

    Sections that are not visible should not be rendered, neither when calling
    model_card.render, nor when calling model_card.save.

    """

    @pytest.fixture
    def template(self):
        return {
            "Model": "Here goes model related stuff",
            "Model/Metrics": "123",
            "Model/Bar": "Baz",
            "Authors": "Jane Doe",
        }

    @pytest.fixture
    def card(self, template):
        model = fit_model()
        card = Card(model, template=template)
        return card

    def test_all_visible_by_default(self, card):
        rendered = card.render()
        expected = (
            "# Model\n\n"
            "Here goes model related stuff\n\n"
            "## Metrics\n\n"
            "123\n\n"
            "## Bar\n\n"
            "Baz\n\n"
            "# Authors\n\n"
            "Jane Doe"
        )
        assert rendered.strip() == expected

    def test_section_invisible(self, card):
        card.select("Model/Metrics").visible = False
        rendered = card.render()
        expected = (
            "# Model\n\n"
            "Here goes model related stuff\n\n"
            "## Bar\n\n"
            "Baz\n\n"
            "# Authors\n\n"
            "Jane Doe"
        )
        assert rendered.strip() == expected

    def test_restoring_visibility_works(self, card):
        card.select("Model/Metrics").visible = False
        card.select("Model/Metrics").visible = True
        expected = (
            "# Model\n\n"
            "Here goes model related stuff\n\n"
            "## Metrics\n\n"
            "123\n\n"
            "## Bar\n\n"
            "Baz\n\n"
            "# Authors\n\n"
            "Jane Doe"
        )
        rendered = card.render()
        assert rendered.strip() == expected

    def test_invisible_parent_section_hides_subsections(self, card):
        # By making the parent section "Model" invisible, all of the subsections
        # are also turned invisible
        card.select("Model").visible = False
        # fmt: off
        expected = (
            "# Authors\n\n"
            "Jane Doe"
        )
        # fmt: on
        rendered = card.render()
        assert rendered.strip() == expected

    def test_visibility_with_card_save(self, card):
        # Since .save and .render share the same functionality, it's not
        # necessary to repeat all the tests above with .save. Just do one test
        # to ensure that the same functionality is indeed being used.
        file = tempfile.mkstemp(suffix=".md", prefix="skops-model-card")[1]
        card.select("Model/Metrics").visible = False
        card.save(file)

        with open(file, "r") as f:
            loaded = f.read()

        expected = (
            "# Model\n\n"
            "Here goes model related stuff\n\n"
            "## Bar\n\n"
            "Baz\n\n"
            "# Authors\n\n"
            "Jane Doe"
        )
        assert loaded.strip() == expected


class TestAddFairlearnMetricFrame:
    @pytest.fixture
    def card(self):
        model = LinearRegression()
        card = Card(model=model)
        return card

    @pytest.fixture
    def metric_frame(self):
        metrics = import_or_raise(
            "fairlearn.metrics", "model card fairlearn metricframe"
        )

        y_true = [1, 1, 1, 1, 1, 0, 0, 1, 1, 0]
        y_pred = [0, 1, 1, 1, 1, 0, 0, 0, 1, 1]
        sex = ["Female"] * 5 + ["Male"] * 5
        metric_dict = {"selection_rate": metrics.selection_rate}
        metric_frame = metrics.MetricFrame(
            y_true=y_true, y_pred=y_pred, sensitive_features=sex, metrics=metric_dict
        )
        return metric_frame

    @pytest.mark.parametrize("transpose", [True, False])
    def test_metric_table(self, card: Card, transpose, metric_frame):
        card.add_fairlearn_metric_frame(
            metric_frame=metric_frame,
            transpose=transpose,
            table_name="Metric Frame Table",
        )

        actual_table = card.select("Metric Frame Table").format()

        if transpose is True:
            expected_table = (
                "<details>\n<summary> Click to expand </summary>\n\n|   selection_rate"
                " |\n|------------------|\n|              0.4 |\n|              0.8"
                " |\n|              0.4 |\n|              0.5 |\n\n</details>"
            )
        else:
            expected_table = (
                "<details>\n<summary> Click to expand </summary>\n\n|   difference |  "
                " group_max |   group_min |   ratio"
                " |\n|--------------|-------------|-------------|---------|\n|         "
                " 0.4 |         0.8 |         0.4 |     0.5 |\n\n</details>"
            )

        assert expected_table == actual_table

    def test_metric_table_with_description(self, card: Card, metric_frame):
        card.add_fairlearn_metric_frame(
            description="An awesome table",
            metric_frame=metric_frame,
            table_name="Metric Frame Table",
        )

        actual_table = card.select("Metric Frame Table").format()
        expected_table = (
            "An awesome table\n\n"
            "<details>\n<summary> Click to expand </summary>\n\n|   selection_rate"
            " |\n|------------------|\n|              0.4 |\n|              0.8"
            " |\n|              0.4 |\n|              0.5 |\n\n</details>"
        )
        assert expected_table == actual_table


class TestCardTableOfContents:
    @pytest.fixture
    def card(self):
        model = LinearRegression()
        card = Card(model=model)
        card.add_model_plot()
        card.add_hyperparams()
        card.add_metrics(accuracy=0.1)
        card.add_get_started_code()
        return card

    def test_toc(self, card):
        toc = card.get_toc()
        exptected_toc = [
            "- Model description",
            "  - Intended uses & limitations",
            "  - Training Procedure",
            "    - Hyperparameters",
            "    - Model Plot",
            "  - Evaluation Results",
            "- How to Get Started with the Model",
            "- Model Card Authors",
            "- Model Card Contact",
            "- Citation",
        ]

        assert toc == "\n".join(exptected_toc)

    def test_toc_with_invisible_section(self, card):
        section = card.select("Citation")
        section.visible = False
        toc = card.get_toc()
        exptected_toc = [
            "- Model description",
            "  - Intended uses & limitations",
            "  - Training Procedure",
            "    - Hyperparameters",
            "    - Model Plot",
            "  - Evaluation Results",
            "- How to Get Started with the Model",
            "- Model Card Authors",
            "- Model Card Contact",
        ]

        assert toc == "\n".join(exptected_toc)



import numpy as np
import pytest
import scipy
import sklearn.tree

from skops.io._utils import get_type_name, get_type_paths


class UserDefinedClass:
    pass


class UserDefinedString(str):
    """Used to test behaviour of subclasses of strings"""

    pass


class TestGetTypeName:
    @pytest.mark.parametrize(
        "input_type, expected_output",
        [
            # Built-In types
            (list, "builtins.list"),
            (set, "builtins.set"),
            (dict, "builtins.dict"),
            (str, "builtins.str"),
            # Numpy types
            (np.ndarray, "numpy.ndarray"),
            (np.ma.MaskedArray, "numpy.ma.core.MaskedArray"),
            # SciPy types
            (scipy.fft.fft, "scipy.fft._basic.fft"),
            # SKlearn types
            (
                sklearn.linear_model.HuberRegressor,
                "sklearn.linear_model._huber.HuberRegressor",
            ),
            # User defined types
            (UserDefinedClass, "test_utils.UserDefinedClass"),
            (UserDefinedString, "test_utils.UserDefinedString"),
        ],
    )
    def test_for_input_types_returns_as_expected(self, input_type, expected_output):
        assert get_type_name(input_type) == expected_output


class TestConvertTypesToStrings:
    @pytest.mark.parametrize(
        "input_list, output_list",
        [
            # Happy path
            (["builtins.str", "builtins.list"], ["builtins.str", "builtins.list"]),
            ([str, list], ["builtins.str", "builtins.list"]),
            ([np.ndarray, "builtins.str"], ["numpy.ndarray", "builtins.str"]),
            # Edge cases
            (None, []),
            (int, ["builtins.int"]),
            ((list,), ["builtins.list"]),
            ([], []),
            (UserDefinedString, ["test_utils.UserDefinedString"]),
            (UserDefinedString("foo"), ["foo"]),
        ],
        ids=[
            "As strings",
            "As types",
            "mixed",
            "None",
            "Single int type",
            "List in tuple",
            "Empty list",
            "UserDefinedString as type",
            "UserDefinedString as instance",
        ],
    )
    def test_for_normal_input_lists_returns_as_expected(self, input_list, output_list):
        assert get_type_paths(input_list) == output_list



from __future__ import annotations

import io
import json
import sys
import warnings
from zipfile import ZipFile

import numpy as np
from scipy import sparse
from sklearn.base import BaseEstimator
from sklearn.utils._testing import assert_allclose_dense_sparse

# TODO: Investigate why that seems to be an issue on MacOS (only observed with
# Python 3.8)
ATOL = 1e-6 if sys.platform == "darwin" else 1e-7


def _is_steps_like(obj):
    # helper function to check if an object is something like Pipeline.steps,
    # i.e. a list of tuples of names and estimators
    if not isinstance(obj, list):  # must be a list
        return False

    if not obj:  # must not be empty
        return False

    if not isinstance(obj[0], tuple):  # must be list of tuples
        return False

    lens = set(map(len, obj))
    if not lens == {2}:  # all elements must be length 2 tuples
        return False

    keys, vals = list(zip(*obj))

    if len(keys) != len(set(keys)):  # keys must be unique
        return False

    if not all(map(lambda x: isinstance(x, (type(None), BaseEstimator)), vals)):
        # values must be BaseEstimators or None
        return False

    return True


def _assert_generic_objects_equal(val1, val2):
    def _is_builtin(val):
        # Check if value is a builtin type
        return getattr(getattr(val, "__class__", {}), "__module__", None) == "builtins"

    if isinstance(val1, (list, tuple, np.ndarray)):
        assert len(val1) == len(val2)
        for subval1, subval2 in zip(val1, val2):
            _assert_generic_objects_equal(subval1, subval2)
            return

    assert type(val1) == type(val2)
    if hasattr(val1, "__dict__"):
        assert_params_equal(val1.__dict__, val2.__dict__)
    elif _is_builtin(val1):
        assert val1 == val2
    else:
        # not a normal Python class, could be e.g. a Cython class
        assert val1.__reduce__() == val2.__reduce__()


def _assert_tuples_equal(val1, val2):
    assert len(val1) == len(val2)
    for subval1, subval2 in zip(val1, val2):
        _assert_vals_equal(subval1, subval2)


def _assert_vals_equal(val1, val2):
    if type(val1) == type:  # e.g. could be np.int64
        assert val1 is val2
    elif hasattr(val1, "__getstate__") and (val1.__getstate__() is not None):
        # This includes BaseEstimator since they implement __getstate__ and
        # that returns the parameters as well.
        # Since Python 3.11, all objects have a __getstate__ but they return
        # None by default, in which case this check is not performed.
        # Some objects return a tuple of parameters, others a dict.
        state1 = val1.__getstate__()
        state2 = val2.__getstate__()
        assert type(state1) == type(state2)
        if isinstance(state1, tuple):
            _assert_tuples_equal(state1, state2)
        else:
            assert_params_equal(val1.__getstate__(), val2.__getstate__())
    elif sparse.issparse(val1):
        assert sparse.issparse(val2) and ((val1 - val2).nnz == 0)
    elif isinstance(val1, (np.ndarray, np.generic)):
        if len(val1.dtype) == 0:
            # for arrays with at least 2 dimensions, check that contiguity is
            # preserved
            if val1.squeeze().ndim > 1:
                assert val1.flags["C_CONTIGUOUS"] is val2.flags["C_CONTIGUOUS"]
                assert val1.flags["F_CONTIGUOUS"] is val2.flags["F_CONTIGUOUS"]
            if val1.dtype == object:
                assert val2.dtype == object
                assert val1.shape == val2.shape
                for subval1, subval2 in zip(val1, val2):
                    _assert_generic_objects_equal(subval1, subval2)
            else:
                # simple comparison of arrays with simple dtypes, almost all
                # arrays are of this sort.
                np.testing.assert_array_equal(val1, val2)
        elif len(val1.shape) == 1:
            # comparing arrays with structured dtypes, but they have to be 1D
            # arrays. This is what we get from the Tree's state.
            assert np.all([x == y for x, y in zip(val1, val2)])
        else:
            # we don't know what to do with these values, for now.
            assert False
    elif isinstance(val1, (tuple, list)):
        assert len(val1) == len(val2)
        for subval1, subval2 in zip(val1, val2):
            _assert_vals_equal(subval1, subval2)
    elif isinstance(val1, float) and np.isnan(val1):
        assert np.isnan(val2)
    elif isinstance(val1, dict):
        # dictionaries are compared by comparing their values recursively.
        assert set(val1.keys()) == set(val2.keys())
        for key in val1:
            _assert_vals_equal(val1[key], val2[key])
    elif hasattr(val1, "__dict__") and hasattr(val2, "__dict__"):
        _assert_vals_equal(val1.__dict__, val2.__dict__)
    elif isinstance(val1, np.ufunc):
        assert val1 == val2
    elif val1.__class__.__module__ == "builtins":
        assert val1 == val2
    else:
        _assert_generic_objects_equal(val1, val2)


def assert_params_equal(params1, params2):
    # helper function to compare estimator dictionaries of parameters
    if params1 is None and params2 is None:
        return
    assert len(params1) == len(params2)
    assert set(params1.keys()) == set(params2.keys())
    for key in params1:
        with warnings.catch_warnings():
            # this is to silence the deprecation warning from _DictWithDeprecatedKeys
            warnings.filterwarnings("ignore", category=FutureWarning, module="sklearn")
            val1, val2 = params1[key], params2[key]
        assert type(val1) == type(val2)

        if _is_steps_like(val1):
            # Deal with Pipeline.steps, FeatureUnion.transformer_list, etc.
            assert _is_steps_like(val2)
            val1, val2 = dict(val1), dict(val2)

        if isinstance(val1, (tuple, list)):
            assert len(val1) == len(val2)
            for subval1, subval2 in zip(val1, val2):
                _assert_vals_equal(subval1, subval2)
        elif isinstance(val1, dict):
            assert_params_equal(val1, val2)
        else:
            _assert_vals_equal(val1, val2)


def assert_method_outputs_equal(estimator, loaded, X):
    # helper function that checks the output of all supported methods
    for method in [
        "predict",
        "predict_proba",
        "decision_function",
        "transform",
        "predict_log_proba",
    ]:
        err_msg = (
            f"{estimator.__class__.__name__}.{method}() doesn't produce the same"
            " results after loading the persisted model."
        )
        if hasattr(estimator, method):
            X_out1 = getattr(estimator, method)(X)
            X_out2 = getattr(loaded, method)(X)
            assert_allclose_dense_sparse(X_out1, X_out2, err_msg=err_msg, atol=ATOL)


def downgrade_state(*, data: bytes, keys: list[str], old_state: dict, protocol: int):
    """Function to downgrade the persisted state of a skops object.

    This function is important for testing upgrades to the skops persistence
    protocol. When an upgrade is made, we add a test to ensure that the old
    state can still be loaded successfully. For this, we need to generate a
    state that looks like it came from the previous protocol. This function
    helps doing that.

    The caller should pass the new state, the path to the sub-state that needs
    to be downgraded, the actual old state to be downgraded to, and the protocol
    of that old version. Then this function will replace the new state with the
    old state and insert the old protocol number. It also adds an ``__id__``
    field, which is expected for memoization.

    Here is an example of how to use it:

    .. code:: python

        estimator = ...
        # get the state of the object using current protocol
        dumped = sio.dumps(estimator)
        # let's assume that estimator.foo.bar was changed
        keys = ["foo", "bar"]
        old_state = old_get_state_function(bar)
        downgraded = downgrad_state(
            data=dumped,
            keys=keys,
            old_state=old_state
            protocol=current_protocol - 1,
        )
        # check that this does not raise an error:
        sio.loads(downgrade, trusted=...)

    Parameters
    ----------
    data : bytes
        The old state, as generated by ``skops.io.dumps``.

    keys : list of str, or None
        The keys that lead to the old state. E.g. if we want to replace
        ``state["foo"]["bar"]``, then keys should be ``["foo", "bar"]``. If
        ``keys=None``, the whole schema is instead replaced by the old state
        being passed.

    old_state : dict
        The old state, as would be produced by the old ``get_state`` function.

    protocol : int
        The protocol number corresponding to the old state.

    Returns
    -------
    bytes
        The old state, as would have been dumped by ``skops.io.dumps``.

    """
    # load from bytes
    with ZipFile(io.BytesIO(data), "r") as zip_file:
        schema = json.loads(zip_file.read("schema.json"))

    # replace state in schema using old state
    if keys is None:
        # replace all fields
        schema = old_state
        schema["__id__"] = id(schema)
    else:
        # replace specific field
        state = schema
        for key in keys[:-1]:
            state = state[key]
        state[keys[-1]] = old_state

        # there has to be an __id__ field for memoization
        state[keys[-1]]["__id__"] = id(schema)

    schema["protocol"] = protocol

    # dump into bytes
    buffer = io.BytesIO()
    with ZipFile(buffer, "w") as zip_file:
        zip_file.writestr("schema.json", json.dumps(schema, indent=2))
    return buffer.getbuffer().tobytes()



from __future__ import annotations

import io
from typing import Any, Sequence

from scipy.sparse import load_npz, save_npz, spmatrix

from ._audit import Node
from ._protocol import PROTOCOL
from ._utils import LoadContext, SaveContext, get_module


def sparse_matrix_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    res = {
        "__class__": obj.__class__.__name__,
        "__module__": get_module(type(obj)),
        "__loader__": "SparseMatrixNode",
    }

    data_buffer = io.BytesIO()
    save_npz(data_buffer, obj)
    # Memoize the object and then check if it's file name (containing
    # the object id) already exists. If it does, there is no need to
    # save the object again. Memoizitation is necessary since for
    # ephemeral objects, the same id might otherwise be reused.
    obj_id = save_context.memoize(obj)
    f_name = f"{obj_id}.npz"
    if f_name not in save_context.zip_file.namelist():
        save_context.zip_file.writestr(f_name, data_buffer.getbuffer())

    res["type"] = "scipy"
    res["file"] = f_name
    return res


class SparseMatrixNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.type = state["type"]
        self.trusted = self._get_trusted(trusted, [spmatrix])
        if self.type != "scipy":
            raise TypeError(
                f"Cannot load object of type {self.module_name}.{self.class_name}"
            )

        self.children = {"content": io.BytesIO(load_context.src.read(state["file"]))}

    def _construct(self):
        # scipy load_npz uses numpy.save with allow_pickle=False under the
        # hood, so we're safe using it
        return load_npz(self.children["content"])


# tuples of type and function that gets the state of that type
GET_STATE_DISPATCH_FUNCTIONS = [
    # use 'spmatrix' to check if a matrix is a sparse matrix because that is
    # what scipy.sparse.issparse checks
    (spmatrix, sparse_matrix_get_state),
]
# tuples of type and function that creates the instance of that type
NODE_TYPE_MAPPING = {
    # use 'spmatrix' to check if a matrix is a sparse matrix because that is
    # what scipy.sparse.issparse checks
    ("SparseMatrixNode", PROTOCOL): SparseMatrixNode,
}



"""Contains the PandocParser

This class needs to know about the pandoc parse tree but should not have
knowledge of any particular markup syntex; everything related to markup should
be known by the mapping attribute.

"""

from __future__ import annotations

import json
import subprocess
from pathlib import Path
from tempfile import mkdtemp
from typing import Any, Literal

import yaml  # type: ignore
from packaging.version import Version

from skops.card import Card
from skops.card._model_card import Section

from ._markup import Markdown, PandocItem

PANDOC_MIN_VERSION = "2.0"


class PandocParser:
    """Create model cards from files parsed through pandoc.

    This class knows about the implementation details of the
    :class:`~skops.card.Card` and generates it by initializing an empty class
    and then calling its methods with the input provided by pandoc.

    ``PandocParser`` does not know about any specific markup type, such as
    markdown. Instead, it is initialized with a ``Mapping``, which is
    responsible to convert pandoc input into the desired markup language.

    After initializing this class, call
    :meth:`~skops.card._parser.PandocParser.generate` to generate the resulting
    :class:`~skops.card.Card` instance.

    Parameters
    ----------
    source : str
        The model card parsed using the ``pandoc -t json`` option.

    markup_type : "markdown"
        The type of markup that was used for the model card. Right now, only
        ``"markdown"`` is supported.

    """

    def __init__(
        self, source: str, markup_type: Literal["markdown"] = "markdown"
    ) -> None:
        self.source = source
        if markup_type.lower() == "markdown":
            self.mapping = Markdown()
        else:
            raise ValueError(f"Markup of type {markup_type} is not supported (yet)")

    def _add_section(
        self, section_name: str, card: Card, section_trace: list[str]
    ) -> Section:
        # Add a new section to the card, which can be a subsection, and return
        # it.
        section_name = "/".join(section_trace)
        cur_section = card._add_single(section_name, "")
        return cur_section

    def _add_content(self, content: str, section: Section | None) -> None:
        # Add content to the current section
        if section is None:
            # This may occur if the model card starts without a section. This is
            # not illegal in markdown, but we don't handle it yet.
            raise ValueError(
                "Trying to add content but there is no current section, "
                "this is probably a bug, please open an issue on GitHub."
            )

        if not section.content:
            section.content = content
        elif isinstance(section.content, str):
            section.content = section.content + "\n\n" + content
        else:  # pragma: no cover
            # TODO: Content is a Formattable, no generic way to modify it --
            # should we require each Formattable to have an update method?
            raise ValueError(f"Could not modify content of {section.content}")

    def _parse_header(
        self, item: PandocItem, section_trace: list[str]
    ) -> tuple[str, int]:
        # Headers are the only type of item that needs to be handled
        # differently. This is because we structure the underlying model card
        # data as a tree with nodes corresponding to headers. To assign the
        # right parent or child node, we need to keep track of the level of the
        # headers. This cannot be done on the level of the markdown mapping,
        # since it is not aware of the tree structure.
        level, _, _ = item["c"]
        content = self.mapping(item)
        return content, level

    def _post_process(self, res: str) -> str:
        # replace Latin1 space
        res = res.replace("\xa0", " ")

        # pandoc creates ☒ and ☐ for to do items but GitHub requires [x] and [ ]
        # for an item to be considered a to do item
        res = res.replace("- ☒", "- [x]").replace("- ☐", "- [ ]")
        return res

    def generate(self) -> Card:
        """Generate the model card instance from the parsed card.

        Returns
        -------
        card : :class:`~skops.card.Card`
            The parsed model card instance. If not further modified, the output
            of saving that card should be (almost) identical to the initial
            model card.
        """
        section: Section | None = None
        section_trace: list[str] = []
        card = Card(None, template=None)

        # Parsing the flat structure, not recursively as in pandocfilters. After
        # visiting the parent node, it's not necessary to visit its child nodes,
        # because the mapping class already takes care of visiting the child
        # nodes.
        for item in json.loads(self.source)["blocks"]:
            if item["t"] == "Header":
                content, level = self._parse_header(item, section_trace=section_trace)
                res = self._post_process(content)
                section_trace = section_trace[: level - 1] + [res]
                section = self._add_section(res, card=card, section_trace=section_trace)
            else:
                res = self._post_process(self.mapping(item))
                self._add_content(res, section=section)

        return card


def _get_pandoc_version() -> str:
    """Shell out to retrieve the pandoc version

    Raises
    ------
    RuntimeError
        If pandoc version could not be determined, raise a ``RuntimeError``.

    Returns
    -------
    pandoc_version : str
        The pandoc version as a list of ints.
    """
    proc = subprocess.run(
        ["pandoc", "--version"],
        capture_output=True,
    )
    version_info = str(proc.stdout.decode("utf-8")).split("\n", 1)[0]
    if not version_info.startswith("pandoc "):
        # pandoc is installed but version cannot be determined
        raise RuntimeError("Could not determine version of pandoc.")

    _, _, pandoc_version = version_info.partition(" ")
    return pandoc_version


def check_pandoc_installed(
    min_version: str | None = PANDOC_MIN_VERSION,
) -> None:
    """Check if pandoc is installed on the system

    Parameters
    ----------
    min_version : list[int] or None
        If passed, check that the pandoc version is greater or equal to this one.

    Raises
    ------
    FileNotFoundError
        When the binary is not found, raise this error.

    RuntimeError
        If pandoc version could not be determined, raise a ``RuntimeError``.

    ValueError
        If min version is passed and not matched or exceeded, raise a ``ValueError``.
    """
    try:
        pandoc_version = _get_pandoc_version()
    except FileNotFoundError as exc:
        # pandoc is not installed
        msg = (
            "This feature requires the pandoc library to be installed on your system, "
            "please follow these install instructions: "
            "https://pandoc.org/installing.html."
        )
        raise FileNotFoundError(msg) from exc

    if not min_version:
        return

    if Version(pandoc_version) < Version(min_version):
        raise ValueError(
            f"Pandoc version too low, expected at least {min_version}, "
            f"got {pandoc_version} instead."
        )


def _card_with_detached_metainfo(path: str | Path) -> tuple[str | Path, dict[str, Any]]:
    """Detach the possibly existing yaml part of the model card

    Model cards always have a markdown part and optionally a yaml part at the
    head, delimited by "---". Obviously, pandoc cannot parse that. Therefore, we
    detach the yaml part and return it as a separate dict, only leaving
    (hopefully) valid markdown.

    path : str or pathlib.Path
        The path to the model card file.

    Returns
    -------
    file : path
        The path to the model card without any yaml metainfo. If the model card
        didn't contain that metainfo to begin with, this is just the path to the
        original model card. If it did contain metainfo, this is a path to a new
        temporary file with the metainfo removed.

    metainfo : dict
        The metainfo from the yaml part as a parsed dict. If no metainfo was
        present, the dict is empty.
    """
    with open(path, "r") as f:
        text = f.read()

    sep_start, sep_end = "---\n", "\n---"

    metainfo: dict[str, Any] = {}
    if not text.startswith(sep_start):  # no metainfo:
        return path, metainfo

    idx_separator = text.find(sep_end)
    if idx_separator < len(sep_start):  # pragma: no cover
        # separator shouldn't come earlier than this
        return path, metainfo

    # https://black.readthedocs.io/en/stable/faq.html#why-are-flake8-s-e203-and-w503-violated
    text_clean = text[idx_separator + len(sep_end) :]  # noqa: E203
    metainfo = yaml.safe_load(  # type: ignore
        text[len(sep_start) : idx_separator]  # noqa: E203
    )

    file = Path(mkdtemp()) / "tmp-model-card.md"
    with open(file, "w") as f:
        f.write(text_clean)
    return file, metainfo


def parse_modelcard(path: str | Path) -> Card:
    """Read a model card and return a Card object

    This allows users to load a dumped model card and continue to edit it.

    Using this function requires ``pandoc`` to be installed. Please follow these
    instructions:

    https://pandoc.org/installing.html

    Examples
    --------
    >>> import numpy as np
    >>> from sklearn.linear_model import LinearRegression
    >>> from skops.card import Card
    >>> from skops.card import parse_modelcard
    >>> X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
    >>> y = np.dot(X, np.array([1, 2])) + 3
    >>> regr = LinearRegression().fit(X, y)
    >>> card = Card(regr)
    >>> card.save("README.md")  # doctest: +SKIP
    >>> # later, load the card again
    >>> parsed_card = parse_modelcard("README.md")  # doctest: +SKIP
    >>> # continue editing the card
    >>> parsed_card.add(**{"My new section": "My new content"})  # doctest: +SKIP
    Card(...)
    >>> # overwrite old card with new one
    >>> parsed_card.save("README.md")  # doctest: +SKIP

    Notes
    -----
    There are some **known limitations** to the parser that may result in the
    model card generated from the parsed file not being 100% identical to the
    original model card:

    - In markdown, bold and italic text can be encoded in different fashions,
      e.g. ``_like this_`` or ``*like this*`` for italic text. Pandoc doesn't
      differentiate between the two. Therefore, the output may use one method
      where the original card used the other. When rendered, the two results
      should, however, be the same.
    - Table alignment may be different. At the moment, skops does not make use
      of column alignment information in tables, so that may differ.
    - Quote symbols may differ, e.g. ``it’s`` becoming ``it's``.
    - The number of empty lines may differ, e.g. two empty lines being
      transformed into one empty line.
    - The optional title of links is not preserved, as e.g. in
      `[text](https://example.com "this disappears")`
    - Trailing whitespace is removed.
    - Tab indentation may be removed, e.g. in raw html.
    - The yaml part of the model card can have some non-semantic differences,
      like omitting optional quotation marks.

    For these reasons, please don't expect the output of a parsed card to be
    100% identical to the original input. However, none of the listed changes
    makes any _semantic_ difference. If you find that there is a semantic
    difference in the output, please open an issue on GitHub.

    Parameters
    ----------
    path : str or pathlib.Path
        The path to the existing model card.

    Returns
    -------
    card : skops.card.Card
        The model card object.

    """
    check_pandoc_installed()

    path, metainfo = _card_with_detached_metainfo(path)

    proc = subprocess.run(
        ["pandoc", "-t", "json", "-s", str(path)],
        capture_output=True,
    )
    source = str(proc.stdout.decode("utf-8"))

    parser = PandocParser(source)
    card = parser.generate()
    for key, val in metainfo.items():
        setattr(card.metadata, key, val)

    return card



"""Test persistence of "external" packages

Packages that are not builtins, standard lib, numpy, scipy, or scikit-learn.

"""

import pytest
from sklearn.datasets import make_classification, make_regression

from skops.io import dumps, loads
from skops.io.tests._utils import assert_method_outputs_equal, assert_params_equal

# Default settings for generated data
N_SAMPLES = 30
N_FEATURES = 10
N_CLASSES = 4  # for classification only


@pytest.fixture(scope="module")
def clf_data():
    X, y = make_classification(
        n_samples=N_SAMPLES,
        n_classes=N_CLASSES,
        n_features=N_FEATURES,
        random_state=0,
        n_redundant=1,
        n_informative=N_FEATURES - 1,
    )
    return X, y


@pytest.fixture(scope="module")
def regr_data():
    X, y = make_regression(n_samples=N_SAMPLES, n_features=N_FEATURES, random_state=0)
    return X, y


@pytest.fixture(scope="module")
def rank_data(clf_data):
    X, y = clf_data
    group = [10 for _ in range(N_SAMPLES // 10)]
    n = sum(group)
    if N_SAMPLES > n:
        group[-1] += N_SAMPLES - n
    assert sum(group) == N_SAMPLES
    return X, y, group


class TestLightGBM:
    """Tests for LGBMClassifier, LGBMRegressor, LGBMRanker"""

    @pytest.fixture(autouse=True)
    def lgbm(self):
        lgbm = pytest.importorskip("lightgbm")
        return lgbm

    @pytest.fixture
    def trusted(self):
        # TODO: adjust once more types are trusted by default
        return [
            "collections.defaultdict",
            "lightgbm.basic.Booster",
            "lightgbm.sklearn.LGBMClassifier",
            "lightgbm.sklearn.LGBMRegressor",
            "lightgbm.sklearn.LGBMRanker",
            "numpy.int32",
            "numpy.int64",
            "sklearn.preprocessing._label.LabelEncoder",
        ]

    boosting_types = ["gbdt", "dart", "goss", "rf"]

    @pytest.mark.parametrize("boosting_type", boosting_types)
    def test_classifier(self, lgbm, clf_data, trusted, boosting_type):
        kw = {}
        if boosting_type == "rf":
            kw["bagging_fraction"] = 0.5
            kw["bagging_freq"] = 2

        estimator = lgbm.LGBMClassifier(boosting_type=boosting_type, **kw)
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_params_equal(estimator.get_params(), loaded.get_params())

        X, y = clf_data
        estimator.fit(X, y)
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_method_outputs_equal(estimator, loaded, X)

    @pytest.mark.parametrize("boosting_type", boosting_types)
    def test_regressor(self, lgbm, regr_data, trusted, boosting_type):
        kw = {}
        if boosting_type == "rf":
            kw["bagging_fraction"] = 0.5
            kw["bagging_freq"] = 2

        estimator = lgbm.LGBMRegressor(boosting_type=boosting_type, **kw)
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_params_equal(estimator.get_params(), loaded.get_params())

        X, y = regr_data
        estimator.fit(X, y)
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_method_outputs_equal(estimator, loaded, X)

    @pytest.mark.parametrize("boosting_type", boosting_types)
    def test_ranker(self, lgbm, rank_data, trusted, boosting_type):
        kw = {}
        if boosting_type == "rf":
            kw["bagging_fraction"] = 0.5
            kw["bagging_freq"] = 2

        estimator = lgbm.LGBMRanker(boosting_type=boosting_type, **kw)
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_params_equal(estimator.get_params(), loaded.get_params())

        X, y, group = rank_data
        estimator.fit(X, y, group=group)
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_method_outputs_equal(estimator, loaded, X)


class TestXGBoost:
    """Tests for XGBClassifier, XGBRegressor, XGBRFClassifier, XGBRFRegressor, XGBRanker

    Known bugs:

    - When initialzing with tree_method=None, its value resolves to "exact", but
      after loading, it resolves to "auto" when calling get_params().
    - When initializing with tree_method='gpu_hist' and gpu_id=None, the
      latter's value resolves to 0, but after loading, it resolves to -1, when
      calling get_params()

    These discrepancies occur regardless of skops, so they're a problem in
    xgboost itself. We assume that this has no practical consequences and thus
    avoid testing these cases. See https://github.com/dmlc/xgboost/issues/8596

    """

    @pytest.fixture(autouse=True)
    def xgboost(self):
        xgboost = pytest.importorskip("xgboost")
        return xgboost

    @pytest.fixture
    def trusted(self):
        # TODO: adjust once more types are trusted by default
        return [
            "xgboost.sklearn.XGBClassifier",
            "xgboost.sklearn.XGBRegressor",
            "xgboost.sklearn.XGBRFClassifier",
            "xgboost.sklearn.XGBRFRegressor",
            "xgboost.sklearn.XGBRanker",
            "builtins.bytearray",
            "xgboost.core.Booster",
        ]

    boosters = ["gbtree", "gblinear", "dart"]
    tree_methods = ["approx", "hist", "auto"]

    @pytest.mark.parametrize("booster", boosters)
    @pytest.mark.parametrize("tree_method", tree_methods)
    def test_classifier(self, xgboost, clf_data, trusted, booster, tree_method):
        if (booster == "gblinear") and (tree_method != "approx"):
            # This parameter combination is not supported in XGBoost
            return

        estimator = xgboost.XGBClassifier(booster=booster, tree_method=tree_method)
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_params_equal(estimator.get_params(), loaded.get_params())

        X, y = clf_data
        estimator.fit(X, y)
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_method_outputs_equal(estimator, loaded, X)

    @pytest.mark.parametrize("booster", boosters)
    @pytest.mark.parametrize("tree_method", tree_methods)
    def test_regressor(self, xgboost, regr_data, trusted, booster, tree_method):
        if (booster == "gblinear") and (tree_method != "approx"):
            # This parameter combination is not supported in XGBoost
            return

        estimator = xgboost.XGBRegressor(booster=booster, tree_method=tree_method)
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_params_equal(estimator.get_params(), loaded.get_params())

        X, y = regr_data
        estimator.fit(X, y)
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_method_outputs_equal(estimator, loaded, X)

    @pytest.mark.parametrize("booster", boosters)
    @pytest.mark.parametrize("tree_method", tree_methods)
    def test_rf_classifier(self, xgboost, clf_data, trusted, booster, tree_method):
        if (booster == "gblinear") and (tree_method != "approx"):
            # This parameter combination is not supported in XGBoost
            return

        estimator = xgboost.XGBRFClassifier(booster=booster, tree_method=tree_method)
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_params_equal(estimator.get_params(), loaded.get_params())

        X, y = clf_data
        estimator.fit(X, y)
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_method_outputs_equal(estimator, loaded, X)

    @pytest.mark.parametrize("booster", boosters)
    @pytest.mark.parametrize("tree_method", tree_methods)
    def test_rf_regressor(self, xgboost, regr_data, trusted, booster, tree_method):
        if (booster == "gblinear") and (tree_method != "approx"):
            # This parameter combination is not supported in XGBoost
            return

        estimator = xgboost.XGBRFRegressor(booster=booster, tree_method=tree_method)
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_params_equal(estimator.get_params(), loaded.get_params())

        X, y = regr_data
        estimator.fit(X, y)
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_method_outputs_equal(estimator, loaded, X)

    @pytest.mark.parametrize("booster", boosters)
    @pytest.mark.parametrize("tree_method", tree_methods)
    def test_ranker(self, xgboost, rank_data, trusted, booster, tree_method):
        if (booster == "gblinear") and (tree_method != "approx"):
            # This parameter combination is not supported in XGBoost
            return

        estimator = xgboost.XGBRanker(booster=booster, tree_method=tree_method)
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_params_equal(estimator.get_params(), loaded.get_params())

        X, y, group = rank_data
        estimator.fit(X, y, group=group)
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_method_outputs_equal(estimator, loaded, X)


class TestCatboost:
    """Tests for CatBoostClassifier, CatBoostRegressor, and CatBoostRanker"""

    # CatBoost data is a little different so that it works as categorical data
    @pytest.fixture(scope="module")
    def cb_clf_data(self, clf_data):
        X, y = clf_data
        X = (X - X.min()).astype(int)
        return X, y

    @pytest.fixture(scope="module")
    def cb_regr_data(self, regr_data):
        X, y = regr_data
        X = (X - X.min()).astype(int)
        return X, y

    @pytest.fixture(scope="module")
    def cb_rank_data(self, rank_data):
        X, y, group = rank_data
        X = (X - X.min()).astype(int)
        group_id = sum([[i] * n for i, n in enumerate(group)], [])
        return X, y, group_id

    @pytest.fixture(autouse=True)
    def catboost(self):
        catboost = pytest.importorskip("catboost")
        return catboost

    @pytest.fixture
    def trusted(self):
        # TODO: adjust once more types are trusted by default
        return [
            "builtins.bytes",
            "numpy.float32",
            "numpy.float64",
            "catboost.core.CatBoostClassifier",
            "catboost.core.CatBoostRegressor",
            "catboost.core.CatBoostRanker",
        ]

    boosting_types = ["Ordered", "Plain"]

    @pytest.mark.parametrize("boosting_type", boosting_types)
    def test_classifier(self, catboost, cb_clf_data, trusted, boosting_type):
        estimator = catboost.CatBoostClassifier(
            verbose=False, boosting_type=boosting_type, allow_writing_files=False
        )
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_params_equal(estimator.get_params(), loaded.get_params())

        X, y = cb_clf_data
        estimator.fit(X, y, cat_features=[0, 1])
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_method_outputs_equal(estimator, loaded, X)

    @pytest.mark.parametrize("boosting_type", boosting_types)
    def test_regressor(self, catboost, cb_regr_data, trusted, boosting_type):
        estimator = catboost.CatBoostRegressor(
            verbose=False, boosting_type=boosting_type, allow_writing_files=False
        )
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_params_equal(estimator.get_params(), loaded.get_params())

        X, y = cb_regr_data
        estimator.fit(X, y, cat_features=[0, 1])
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_method_outputs_equal(estimator, loaded, X)

    @pytest.mark.parametrize("boosting_type", boosting_types)
    def test_ranker(self, catboost, cb_rank_data, trusted, boosting_type):
        estimator = catboost.CatBoostRanker(
            verbose=False, boosting_type=boosting_type, allow_writing_files=False
        )
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_params_equal(estimator.get_params(), loaded.get_params())

        X, y, group_id = cb_rank_data
        estimator.fit(X, y, cat_features=[0, 1], group_id=group_id)
        loaded = loads(dumps(estimator), trusted=trusted)
        assert_method_outputs_equal(estimator, loaded, X)



from __future__ import annotations

import io
import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Iterator, Literal
from zipfile import ZipFile

from ._audit import Node, get_tree
from ._general import FunctionNode, JsonNode, ListNode
from ._numpy import NdArrayNode
from ._scipy import SparseMatrixNode
from ._utils import LoadContext


@dataclass
class NodeInfo:
    """Information pertinent for visualizatoin, extracted from ``Node``s.

    This class contains all information necessary for visualizing nodes. This
    way, we can have separate functions for:

    - visiting nodes and determining their safety
    - visualizing the nodes

    The visualization function will only receive the ``NodeInfo`` and does not
    have to concern itself with how to discover children or determine safety.

    """

    level: int
    key: str  # the key to the node
    val: str  # the value of the node
    is_self_safe: bool  # whether this specific node is safe
    is_safe: bool  # whether this node and all of its children are safe
    is_last: bool  # whether this is the last child of parent node


def _check_visibility(
    is_self_safe: bool,
    is_safe: bool,
    show: Literal["all", "untrusted", "trusted"],
) -> bool:
    """Determine visibility of the node.

    Users can indicate if they want to see all nodes, all trusted nodes, or all
    untrusted nodes.

    """
    if show == "all":
        return True

    if show == "untrusted":
        return not is_safe

    # case: show only safe node
    return is_self_safe


def _get_node_label(
    node: NodeInfo,
    tag_safe: str = "",
    tag_unsafe: str = "[UNSAFE]",
    use_colors: bool = True,
    # use rich for coloring
    color_safe: str = "green",
    color_unsafe: str = "red",
    color_child_unsafe: str = "yellow",
):
    """Determine the label of a node.

    Nodes are labeled differently based on how they're trusted.

    """
    # note: when changing the arguments to this function, also update the
    # docstring of visualize!

    if use_colors:
        try:
            import rich  # noqa
        except ImportError:
            use_colors = False

    # add tag if necessary
    node_val = node.val
    tag = tag_safe if node.is_self_safe else tag_unsafe
    if tag:
        node_val += f" {tag}"

    # colorize if so desired and if rich is installed
    if use_colors:
        if node.is_safe:
            color = color_safe
        elif node.is_self_safe:
            color = color_child_unsafe
        else:
            color = color_unsafe
        node_val = f"[{color}]{node_val}"

    return node_val


def pretty_print_tree(
    nodes_iter: Iterator[NodeInfo],
    show: Literal["all", "untrusted", "trusted"],
    **kwargs,
) -> None:
    # This function loops through the flattened nodes of the tree and creates a
    # tree visualization based on the node information. If rich is installed,
    # nodes can be colored.

    print_ = print
    try:
        import rich

        # use rich for printing if available
        print_ = rich.print  # type: ignore
    except ImportError:
        pass

    # start with root node
    node = next(nodes_iter)
    label = _get_node_label(node, **kwargs)
    print_(f"{node.key}: {label}")
    prev_level = node.level  # should be 0
    prefix = ""

    for node in nodes_iter:
        visible = _check_visibility(node.is_self_safe, node.is_safe, show=show)
        if not visible:
            continue

        level_diff = prev_level - node.level
        if level_diff < -1:
            # this would mean it is a "(great-)grandchild" node
            raise ValueError(
                "While constructing the tree of the object, a level difference of "
                f"{level_diff} was encountered, which should not be possible, please "
                "report the issue here: https://github.com/skops-dev/skops/issues"
            )

        # Level diff of -1 means that this node is a child of the previous node.
        # E.g. if the current level if 4 and the previous level was 3, the
        # current node is a child node of the previous one. Since the prefix for
        # a child node was already added, there is nothing more left to do.

        for _ in range(level_diff + 1):
            # This loop is entered if the current node is at the same level as,
            # or higher than, the previous node. This means the prefix has to be
            # truncated according to the level difference. E.g. if the current
            # level is 2 and previous level was 3, it means that we should move
            # up 2 layers of nesting, therefore, we trunce 3-2+1 = 2 times.
            prefix = prefix[:-4]

        print_(prefix, end="")
        if node.is_last:
            print_("└──", end="")
            prefix += "    "
        else:
            print_("├──", end="")
            prefix += "│   "

        label = _get_node_label(node, **kwargs)
        print_(f" {node.key}: {label}")

        prev_level = node.level


def walk_tree(
    node: Node | dict[str, Node] | list[Node],
    node_name: str = "root",
    level: int = 0,
    is_last: bool = False,
) -> Iterator[NodeInfo]:
    """Visit all nodes of the tree and yield their important attributes.

    This function visits all nodes of the object tree and determines:

    - level: how nested the node is
    - key: the key of the node. E.g. if the node is an attribute of an object,
      the key would be the name of the attribute.
    - val: the value of the node, e.g. builtins.list
    - safety: whether it, and its children, are trusted

    These values are just yielded in a flat manner. This way, the consumer of
    this function doesn't need to know how nodes can be nested and how safety of
    a node is determined.

    Parameters
    ----------
    node: :class:`skops.io._audit.Node`
        The current node to visit. Children are visited recursively.

    show: "all" or "untrusted" or "trusted"
        Whether to print all nodes, only untrusted nodes, or only trusted nodes.

    node_name: str (default="root")
        The key to the current node. If "key_types" is encountered, it is
        skipped.

    level: int (default=0)
        The current level of nesting.

    is_last: bool (default=False)
        Whether this is the last node among its sibling nodes.

    Yields
    ------
    :class:`~NodeInfo`:
        A dataclass containing the aforementioned information.

    """
    # key_types is not helpful, as it is artificially added by skops to
    # circumvent the fact that json only allows keys to be strings. It is not
    # useful to the user and adds a lot of noise, thus skip key_types.
    if node_name == "key_types":
        if isinstance(node, ListNode) and node.is_safe():
            return
        raise ValueError(
            "An invalid 'key_types' node was encountered, please report the issue "
            "here: https://github.com/skops-dev/skops/issues"
        )

    if isinstance(node, dict):
        num_nodes = len(node)
        for i, (key, val) in enumerate(node.items(), start=1):
            yield from walk_tree(
                val,
                node_name=key,
                level=level,
                is_last=i == num_nodes,
            )
        return

    if isinstance(node, (list, tuple)):
        num_nodes = len(node)
        for i, val in enumerate(node, start=1):
            yield from walk_tree(
                val,
                node_name=node_name,
                level=level,
                is_last=i == num_nodes,
            )
        return

    # NO MATCH: RAISE ERROR
    if not isinstance(node, Node):
        raise TypeError(
            f"Cannot deal with {type(node)}, please report the issue here: "
            "https://github.com/skops-dev/skops/issues"
        )

    # YIELDING THE ACTUAL NODE INFORMATION HERE

    # Note: calling node.is_safe() on all nodes is potentially wasteful because
    # it is already a recursive call, i.e. child nodes will be checked many
    # times. A solution to this would be to add caching to its call.
    yield NodeInfo(
        level=level,
        key=node_name,
        val=node.format(),
        is_self_safe=node.is_self_safe(),
        is_safe=node.is_safe(),
        is_last=is_last,
    )

    # TYPES WHOSE CHILDREN IT MAKES NO SENSE TO VISIT
    # TODO: For better security, we should check the schema if we return early,
    # otherwise something nefarious could be hidden inside (however, if there
    # is, the node should be marked as unsafe)
    if isinstance(node, (NdArrayNode, SparseMatrixNode, FunctionNode, JsonNode)):
        return

    yield from walk_tree(
        node.children,
        node_name=node_name,
        level=level + 1,
    )


def visualize(
    file: Path | str | bytes,
    show: Literal["all", "untrusted", "trusted"] = "all",
    sink: Callable[..., None] = pretty_print_tree,
    **kwargs: Any,
) -> None:
    """Visualize the contents of a skops file.

    Shows the schema of a skops file as a tree view. In particular, highlights
    untrusted nodes. A node is considered untrusted if at least one of its child
    nodes is untrusted.

    Visualizing the tree using the default visualization function requires the
    ``rich`` library, which can be installed as:

        python -m pip install rich

    If passing a custom visualization function to ``sink``, ``rich`` is not
    required.

    Parameters
    ----------
    file: str or pathlib.Path
        The file name of the object to be loaded.

    show: "all" or "untrusted" or "trusted"
        Whether to print all nodes, only untrusted nodes, or only trusted nodes.

    sink: function (default=:func:`~pretty_print_tree`)

        This function should take at least two arguments, an iterator of
        :class:`~NodeInfo` instances and an indicator of what to show. The
        ``NodeInfo`` contains the information about the node, namely:

            - the level of nesting (int)
            - the key of the node (str)
            - the value of the node as a string representation (str)
            - the safety of the node and its children

        The ``show`` argument is explained above. Any additional ``kwargs``
        passed to ``visualize`` will also be passed to ``sink``.

        The default sink is :func:`~pretty_print_tree`, which takes these
        additional parameters:

            - tag_safe: The tag used to mark trusted nodes (default="", i.e no
              tag)
            - tag_unsafe: The tag used to mark untrusted nodes
              (default="[UNSAFE]")
            - use_colors: Whether to colorize the nodes (default=True). Colors
              requires the ``rich`` package to be installed.
            - color_safe: Color to use for trusted nodes (default="green")
            - color_unsafe: Color to use for untrusted nodes (default="red")
            - color_child_unsafe: Color to use for nodes that are trusted but
              that have untrusted child ndoes (default="yellow")

        So if you don't want to have colored output, just pass
        ``use_colors=False`` to ``visualize``. The colors themselves, such
        as "red" and "green", refer to the standard colors used by ``rich``.

    """
    if isinstance(file, bytes):
        zf = ZipFile(io.BytesIO(file), "r")
    else:
        zf = ZipFile(file, "r")

    with zf as zip_file:
        schema = json.loads(zip_file.read("schema.json"))
        load_context = LoadContext(src=zip_file, protocol=schema["protocol"])
        tree = get_tree(schema, load_context=load_context)

    nodes = walk_tree(tree)
    # TODO: it would be nice to print html representation if inside a notebook
    sink(nodes, show, **kwargs)



from __future__ import annotations

import io
import json
import operator
import uuid
from functools import partial
from types import FunctionType, MethodType
from typing import Any, Sequence

import numpy as np

from ._audit import Node, get_tree
from ._protocol import PROTOCOL
from ._trusted_types import (
    PRIMITIVE_TYPE_NAMES,
    SCIPY_UFUNC_TYPE_NAMES,
    SKLEARN_ESTIMATOR_TYPE_NAMES,
)
from ._utils import (
    LoadContext,
    SaveContext,
    _import_obj,
    get_module,
    get_state,
    gettype,
)
from .exceptions import UnsupportedTypeException


def dict_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    res = {
        "__class__": obj.__class__.__name__,
        "__module__": get_module(type(obj)),
        "__loader__": "DictNode",
    }

    key_types = get_state([type(key) for key in obj.keys()], save_context)
    content = {}
    for key, value in obj.items():
        if isinstance(value, property):
            continue
        if np.isscalar(key) and hasattr(key, "item"):
            # convert numpy value to python object
            key = key.item()  # type: ignore
        content[key] = get_state(value, save_context)
    res["content"] = content
    res["key_types"] = key_types
    return res


class DictNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.trusted = self._get_trusted(trusted, [dict])
        self.children = {
            "key_types": get_tree(state["key_types"], load_context),
            "content": {
                key: get_tree(value, load_context)
                for key, value in state["content"].items()
            },
        }

    def _construct(self):
        content = gettype(self.module_name, self.class_name)()
        key_types = self.children["key_types"].construct()
        for k_type, (key, val) in zip(key_types, self.children["content"].items()):
            content[k_type(key)] = val.construct()
        return content


def list_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    res = {
        "__class__": obj.__class__.__name__,
        "__module__": get_module(type(obj)),
        "__loader__": "ListNode",
    }
    content = [get_state(value, save_context) for value in obj]

    res["content"] = content
    return res


class ListNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.trusted = self._get_trusted(trusted, [list])
        self.children = {
            "content": [get_tree(value, load_context) for value in state["content"]]
        }

    def _construct(self):
        content_type = gettype(self.module_name, self.class_name)
        return content_type([item.construct() for item in self.children["content"]])


def set_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    res = {
        "__class__": obj.__class__.__name__,
        "__module__": get_module(type(obj)),
        "__loader__": "SetNode",
    }
    content = [get_state(value, save_context) for value in obj]
    res["content"] = content
    return res


class SetNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.trusted = self._get_trusted(trusted, [set])
        self.children = {
            "content": [get_tree(value, load_context) for value in state["content"]]
        }

    def _construct(self):
        content_type = gettype(self.module_name, self.class_name)
        return content_type([item.construct() for item in self.children["content"]])


def tuple_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    res = {
        "__class__": obj.__class__.__name__,
        "__module__": get_module(type(obj)),
        "__loader__": "TupleNode",
    }
    content = tuple(get_state(value, save_context) for value in obj)
    res["content"] = content
    return res


class TupleNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.trusted = self._get_trusted(trusted, [tuple])
        self.children = {
            "content": [get_tree(value, load_context) for value in state["content"]]
        }

    def _construct(self):
        # Returns a tuple or a namedtuple instance.

        cls = gettype(self.module_name, self.class_name)
        content = tuple(value.construct() for value in self.children["content"])

        if self.isnamedtuple(cls):
            return cls(*content)
        return content

    def isnamedtuple(self, t) -> bool:
        # This is needed since namedtuples need to have the args when
        # initialized.
        b = t.__bases__
        if len(b) != 1 or b[0] != tuple:
            return False
        f = getattr(t, "_fields", None)
        if not isinstance(f, tuple):
            return False
        return all(type(n) == str for n in f)


def function_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    res = {
        "__class__": obj.__name__,
        "__module__": get_module(obj),
        "__loader__": "FunctionNode",
    }
    return res


class FunctionNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        # TODO: what do we trust?
        self.trusted = self._get_trusted(trusted, default=SCIPY_UFUNC_TYPE_NAMES)
        self.children = {}

    def _construct(self):
        return gettype(self.module_name, self.class_name)

    def _get_function_name(self) -> str:
        return f"{self.module_name}.{self.class_name}"

    def get_unsafe_set(self) -> set[str]:
        fn_name = self._get_function_name()
        if (self.trusted is True) or (fn_name in self.trusted):
            return set()

        return {fn_name}


def partial_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    _, _, (func, args, kwds, namespace) = obj.__reduce__()
    res = {
        "__class__": "partial",  # don't allow any subclass
        "__module__": get_module(type(obj)),
        "__loader__": "PartialNode",
        "content": {
            "func": get_state(func, save_context),
            "args": get_state(args, save_context),
            "kwds": get_state(kwds, save_context),
            "namespace": get_state(namespace, save_context),
        },
    }
    return res


class PartialNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        # TODO: should we trust anything?
        self.trusted = self._get_trusted(trusted, [])
        self.children = {
            "func": get_tree(state["content"]["func"], load_context),
            "args": get_tree(state["content"]["args"], load_context),
            "kwds": get_tree(state["content"]["kwds"], load_context),
            "namespace": get_tree(state["content"]["namespace"], load_context),
        }

    def _construct(self):
        func = self.children["func"].construct()
        args = self.children["args"].construct()
        kwds = self.children["kwds"].construct()
        namespace = self.children["namespace"].construct()
        instance = partial(func, *args, **kwds)  # always use partial, not a subclass
        # partial always has __setstate__
        instance.__setstate__((func, args, kwds, namespace))  # type: ignore
        return instance


def type_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    # To serialize a type, we first need to set the metadata to tell that it's
    # a type, then store the type's info itself in the content field.
    res = {
        "__class__": obj.__name__,
        "__module__": get_module(obj),
        "__loader__": "TypeNode",
    }
    return res


class TypeNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        # TODO: what do we trust?
        self.trusted = self._get_trusted(trusted, PRIMITIVE_TYPE_NAMES)
        # We use a bare Node type here since a Node only checks the type in the
        # dict using __class__ and __module__ keys.
        self.children = {}

    def _construct(self):
        return _import_obj(self.module_name, self.class_name)


def slice_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    res = {
        "__class__": obj.__class__.__name__,
        "__module__": get_module(type(obj)),
        "__loader__": "SliceNode",
        "content": {
            "start": obj.start,
            "stop": obj.stop,
            "step": obj.step,
        },
    }
    return res


class SliceNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.trusted = self._get_trusted(trusted, [slice])
        self.children = {
            "start": state["content"]["start"],
            "stop": state["content"]["stop"],
            "step": state["content"]["step"],
        }

    def _construct(self):
        return slice(
            self.children["start"], self.children["stop"], self.children["step"]
        )

    def get_unsafe_set(self):
        return set()


def object_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    # This method is for objects which can either be persisted with json, or
    # the ones for which we can get/set attributes through
    # __getstate__/__setstate__ or reading/writing to __dict__.
    try:
        # if we can simply use json, then we're done.
        obj_str = json.dumps(obj)
        return {
            "__class__": "str",
            "__module__": "builtins",
            "__loader__": "JsonNode",
            "content": obj_str,
            "is_json": True,
        }
    except Exception:
        pass

    res = {
        "__class__": obj.__class__.__name__,
        "__module__": get_module(type(obj)),
        "__loader__": "ObjectNode",
    }

    # __getstate__ takes priority over __dict__, and if non exist, we only save
    # the type of the object, and loading would mean instantiating the object.
    if hasattr(obj, "__getstate__"):
        attrs = obj.__getstate__()
    elif hasattr(obj, "__dict__"):
        attrs = obj.__dict__
    else:
        return res

    content = get_state(attrs, save_context)
    # it's sufficient to store the "content" because we know that this dict can
    # only have str type keys
    res["content"] = content
    return res


class ObjectNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)

        content = state.get("content")
        if content is not None:
            attrs = get_tree(content, load_context)
        else:
            attrs = None

        self.children = {"attrs": attrs}
        # TODO: what do we trust?
        self.trusted = self._get_trusted(trusted, default=SKLEARN_ESTIMATOR_TYPE_NAMES)

    def _construct(self):
        cls = gettype(self.module_name, self.class_name)

        # Instead of simply constructing the instance, we use __new__, which
        # bypasses the __init__, and then we set the attributes. This solves the
        # issue of required init arguments. Note that the instance created here
        # might not be valid until all its attributes have been set below.
        instance = cls.__new__(cls)  # type: ignore

        if not self.children["attrs"]:
            # nothing more to do
            return instance

        attrs = self.children["attrs"].construct()
        if attrs is not None:
            if hasattr(instance, "__setstate__"):
                instance.__setstate__(attrs)
            else:
                instance.__dict__.update(attrs)

        return instance


def method_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    # This method is used to persist bound methods, which are
    # dependent on a specific instance of an object.
    # It stores the state of the object the method is bound to,
    # and prepares both to be persisted.
    res = {
        "__class__": obj.__class__.__name__,
        "__module__": get_module(obj),
        "__loader__": "MethodNode",
        "content": {
            "func": obj.__func__.__name__,
            "obj": get_state(obj.__self__, save_context),
        },
    }
    return res


class MethodNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.children = {
            "obj": get_tree(state["content"]["obj"], load_context),
            "func": state["content"]["func"],
        }
        # TODO: what do we trust?
        self.trusted = self._get_trusted(trusted, [])

    def _construct(self):
        loaded_obj = self.children["obj"].construct()
        method = getattr(loaded_obj, self.children["func"])
        return method


def unsupported_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    raise UnsupportedTypeException(obj)


class JsonNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.content = state["content"]
        self.children = {}

    def is_safe(self) -> bool:
        # JsonNode is always considered safe.
        # TODO: should we consider a JsonNode always safe?
        return True

    def is_self_safe(self) -> bool:
        return True

    def get_unsafe_set(self) -> set[str]:
        return set()

    def _construct(self):
        return json.loads(self.content)

    def format(self) -> str:
        """Representation of the node's content.

        Since no module is used, just show the content.

        """
        return f"json-type({self.content})"


def bytes_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    f_name = f"{uuid.uuid4()}.bin"
    save_context.zip_file.writestr(f_name, obj)
    res = {
        "__class__": obj.__class__.__name__,
        "__module__": get_module(type(obj)),
        "__loader__": "BytesNode",
        "file": f_name,
    }
    return res


def bytearray_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    res = bytes_get_state(obj, save_context)
    res["__loader__"] = "BytearrayNode"
    return res


class BytesNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.trusted = self._get_trusted(trusted, [bytes])
        self.children = {"content": io.BytesIO(load_context.src.read(state["file"]))}

    def _construct(self):
        content = self.children["content"].getvalue()
        return content


class BytearrayNode(BytesNode):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.trusted = self._get_trusted(trusted, [bytearray])

    def _construct(self):
        content_bytes = super()._construct()
        content_bytearray = bytearray(list(content_bytes))
        return content_bytearray


def operator_func_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    _, attrs = obj.__reduce__()
    res = {
        "__class__": obj.__class__.__name__,
        "__module__": "operator",
        "__loader__": "OperatorFuncNode",
        "attrs": get_state(attrs, save_context),
    }
    return res


class OperatorFuncNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.trusted = self._get_trusted(trusted, [])
        self.children["attrs"] = get_tree(state["attrs"], load_context)

    def _construct(self):
        op = getattr(operator, self.class_name)
        attrs = self.children["attrs"].construct()
        return op(*attrs)


# <class 'builtin_function_or_method'>
builtin_function_or_method = type(len)


# tuples of type and function that gets the state of that type
GET_STATE_DISPATCH_FUNCTIONS = [
    (dict, dict_get_state),
    (list, list_get_state),
    (set, set_get_state),
    (tuple, tuple_get_state),
    (bytes, bytes_get_state),
    (bytearray, bytearray_get_state),
    (slice, slice_get_state),
    (FunctionType, function_get_state),
    (MethodType, method_get_state),
    (partial, partial_get_state),
    (type, type_get_state),
    (builtin_function_or_method, type_get_state),
    (operator.attrgetter, operator_func_get_state),
    (operator.itemgetter, operator_func_get_state),
    (operator.methodcaller, operator_func_get_state),
    (object, object_get_state),
]

NODE_TYPE_MAPPING = {
    ("DictNode", PROTOCOL): DictNode,
    ("ListNode", PROTOCOL): ListNode,
    ("SetNode", PROTOCOL): SetNode,
    ("TupleNode", PROTOCOL): TupleNode,
    ("BytesNode", PROTOCOL): BytesNode,
    ("BytearrayNode", PROTOCOL): BytearrayNode,
    ("SliceNode", PROTOCOL): SliceNode,
    ("FunctionNode", PROTOCOL): FunctionNode,
    ("MethodNode", PROTOCOL): MethodNode,
    ("PartialNode", PROTOCOL): PartialNode,
    ("TypeNode", PROTOCOL): TypeNode,
    ("ObjectNode", PROTOCOL): ObjectNode,
    ("JsonNode", PROTOCOL): JsonNode,
    ("OperatorFuncNode", PROTOCOL): OperatorFuncNode,
}
